# type: ignore

"""Tests for the mycli.clistyle module."""

from types import SimpleNamespace

from prompt_toolkit.styles import Style as PromptStyle
from pygments.style import Style as PygmentsStyle
from pygments.token import Token
from pygments.util import ClassNotFound

from mycli import clistyle


def test_parse_pygments_style_handles_style_classes_instances_and_dict_values() -> None:
    class DemoStyle(PygmentsStyle):
        default_style = ''
        styles = {
            Token.Name: 'bold',
            Token.String: 'ansired',
        }

    token_type, style_value = clistyle.parse_pygments_style(
        'Token.String',
        DemoStyle,
        {'Token.String': 'Token.Name'},
    )
    assert token_type == Token.String
    assert style_value == 'bold'

    token_type, style_value = clistyle.parse_pygments_style(
        'Token.String',
        DemoStyle(),
        {'Token.String': 'Token.Name'},
    )
    assert token_type == Token.String
    assert style_value == 'bold'

    token_type, style_value = clistyle.parse_pygments_style(
        'Token.String',
        'unused',
        {'Token.String': 'ansiblue'},
    )
    assert token_type == Token.String
    assert style_value == 'ansiblue'


def test_is_valid_pygments_returns_true_and_false(monkeypatch) -> None:
    assert clistyle.is_valid_pygments('ansired') is True

    class FailingPygmentsStyle:
        def __init_subclass__(cls, **kwargs) -> None:
            raise AssertionError('bad style')

    monkeypatch.setattr(clistyle, 'PygmentsStyle', FailingPygmentsStyle)

    assert clistyle.is_valid_pygments('invalid') is False


def test_is_valid_ptoolkit_returns_true_and_false(monkeypatch) -> None:
    assert clistyle.is_valid_ptoolkit('bold') is True

    class FailingPromptStyle:
        def __init__(self, _rules) -> None:
            raise ValueError('bad style')

    monkeypatch.setattr(clistyle, 'Style', FailingPromptStyle)

    assert clistyle.is_valid_ptoolkit('invalid') is False


def test_style_factory_ptoolkit_builds_styles_and_falls_back(monkeypatch, caplog) -> None:
    calls: list[str] = []
    native_style = object()

    def fake_get_style_by_name(name: str):
        calls.append(name)
        if name == 'missing':
            raise ClassNotFound('missing')
        if name == 'native':
            return native_style
        raise AssertionError(f'unexpected style {name}')

    class FakeStyle:
        def __init__(self, rules) -> None:
            self.rules = list(rules)

    monkeypatch.setattr(clistyle.pygments.styles, 'get_style_by_name', fake_get_style_by_name)
    monkeypatch.setattr(
        clistyle,
        'parse_pygments_style',
        lambda token, style, cli_style: {
            'Token.Prompt': (Token.Prompt, 'token-valid'),
            'Token.Toolbar': (Token.Toolbar, 'token-invalid'),
            'Token.Name': (Token.Name, 'token-invalid'),
        }[token],
    )
    monkeypatch.setattr(clistyle, 'is_valid_ptoolkit', lambda value: value in {'token-valid', 'prompt-valid'})
    monkeypatch.setattr(clistyle, 'Style', FakeStyle)
    monkeypatch.setattr(clistyle, 'style_from_pygments_cls', lambda style: ('pygments-style', style))
    monkeypatch.setattr(clistyle, 'merge_styles', lambda styles: styles)

    cli_style = {
        'Token.Prompt': 'Token.Name',
        'Token.Toolbar': 'Token.Name',
        'Token.Name': 'ignored',
        'prompt': 'prompt-valid',
        'search': 'prompt-invalid',
    }

    with caplog.at_level('ERROR', logger='mycli.clistyle'):
        styles = clistyle.style_factory_ptoolkit('missing', cli_style)

    assert calls == ['missing', 'native']
    assert styles[0] == ('pygments-style', native_style)
    assert styles[1].rules == [('bottom-toolbar', 'noreverse')]
    assert styles[2].rules == [
        ('prompt', 'token-valid'),
        ('prompt', 'prompt-valid'),
    ]
    assert ('bottom-toolbar', 'token-invalid') not in styles[2].rules
    assert ('search', 'prompt-invalid') not in styles[2].rules
    assert 'Unhandled style / class name: Token.Name' in caplog.text


def test_style_factory_helpers_updates_known_tokens(monkeypatch, caplog) -> None:
    base_styles = {Token.Output.Header: 'ansiyellow'}
    style_class = SimpleNamespace(styles=base_styles)

    monkeypatch.setattr(clistyle.pygments.styles, 'get_style_by_name', lambda name: style_class)
    monkeypatch.setattr(
        clistyle,
        'parse_pygments_style',
        lambda token, style, cli_style: {
            'Token.Prompt': (Token.Prompt, 'ansiblue'),
            'Token.Toolbar': (Token.Toolbar, 'skip-me'),
        }[token],
    )
    monkeypatch.setattr(clistyle, 'is_valid_pygments', lambda value: value != 'skip-me')

    cli_style = {
        'Token.Prompt': 'Token.Name',
        'Token.Toolbar': 'Token.Name',
        'search': 'ansigreen',
        'search.current': 'skip-me',
        'sql.keyword': 'ansired',
        'sql.string': 'skip-me',
        'unknown': 'skip-me',
    }

    with caplog.at_level('ERROR', logger='mycli.clistyle'):
        output_style = clistyle.style_factory_helpers('native', cli_style)

    assert output_style.styles[Token.Prompt] == 'ansiblue'
    assert output_style.styles[Token.SearchMatch] == 'ansigreen'
    assert Token.SearchMatch.Current not in output_style.styles
    assert output_style.styles[Token.Keyword] == 'ansired'
    assert output_style.styles[Token.Output.Header] == 'ansiyellow'
    assert Token.Toolbar not in output_style.styles
    assert output_style.styles[Token.String] != 'skip-me'
    assert 'Unhandled style / class name: unknown' in caplog.text


def test_style_factory_helpers_falls_back_and_copies_warning_styles(monkeypatch) -> None:
    native_styles = {
        Token.Text: 'ansiblack',
        Token.Warnings.Header: 'ansimagenta',
        Token.Warnings.Status: 'ansicyan',
    }

    def fake_get_style_by_name(name: str):
        if name == 'missing':
            raise ClassNotFound('missing')
        if name == 'native':
            return SimpleNamespace(styles=native_styles.copy())
        raise AssertionError(f'unexpected style {name}')

    monkeypatch.setattr(clistyle.pygments.styles, 'get_style_by_name', fake_get_style_by_name)

    output_style = clistyle.style_factory_helpers('missing', {}, warnings=True)

    assert output_style.styles[Token.Warnings.Header] == 'ansimagenta'
    assert output_style.styles[Token.Warnings.Status] == 'ansicyan'
    assert output_style.styles[Token.Output.Header] == 'ansimagenta'
    assert output_style.styles[Token.Output.Status] == 'ansicyan'


def test_style_factory_ptoolkit_returns_merged_style_object() -> None:
    style = clistyle.style_factory_ptoolkit('native', {'prompt': 'bold'})

    assert style.get_attrs_for_style_str('class:prompt') == PromptStyle([('prompt', 'bold')]).get_attrs_for_style_str('class:prompt')



# type: ignore

import builtins
from datetime import time
import importlib.util
import os
from pathlib import Path
import sys
from types import SimpleNamespace

from prompt_toolkit.formatted_text import FormattedText
import pymysql
import pytest

from mycli.constants import TEST_DATABASE
from mycli.packages.special import iocommands
from mycli.packages.sqlresult import SQLResult
import mycli.sqlexecute as sqlexecute
from mycli.sqlexecute import ServerInfo, ServerSpecies, SQLExecute
from test.utils import dbtest, is_expanded_output, run, set_expanded_output


def assert_result_equal(
    result,
    preamble=None,
    header=None,
    rows=None,
    status=None,
    status_plain=None,
    postamble=None,
    auto_status=True,
    assert_contains=False,
):
    """Assert that an sqlexecute.run() result matches the expected values."""
    if status_plain is None and auto_status and rows:
        status_plain = f"{len(rows)} row{'s' if len(rows) > 1 else ''} in set"
        status = FormattedText([('', status_plain)])
    fields = {
        "preamble": preamble,
        "header": header,
        "rows": rows,
        "postamble": postamble,
        "status": status,
        "status_plain": status_plain,
    }

    if assert_contains:
        # Do a loose match on the results using the *in* operator.
        for key, field in fields.items():
            if field:
                assert field in result[0][key]
    else:
        # Do an exact match on the fields.
        assert result == [fields]


@dbtest
def test_timediff_negative_value(executor):
    sql = "select timediff('2020-11-11 01:01:01', '2020-11-11 01:02:01')"
    result = run(executor, sql)
    # negative value comes back as str
    assert result[0]["rows"][0][0] == "-00:01:00"


@dbtest
def test_timediff_positive_value(executor):
    sql = "select timediff('2020-11-11 01:02:01', '2020-11-11 01:01:01')"
    result = run(executor, sql)
    # positive value comes back as datetime.time
    assert result[0]["rows"][0][0] == time(0, 1)


@dbtest
def test_get_result_status_without_warning(executor):
    sql = "select 1"
    result = run(executor, sql)
    assert result[0]["status_plain"] == "1 row in set"


@dbtest
def test_get_result_status_with_warning(executor):
    sql = "SELECT 1 + '0 foo'"
    result = run(executor, sql)
    assert result[0]["status"] == FormattedText([
        ('', '1 row in set'),
        ('', ', '),
        ('class:output.status.warning-count', '1 warning'),
    ])
    assert result[0]["status_plain"] == "1 row in set, 1 warning"


@dbtest
def test_conn(executor):
    run(executor, """create table test(a text)""")
    run(executor, """insert into test values('abc')""")
    results = run(executor, """select * from test""")

    assert_result_equal(results, header=["a"], rows=[("abc",)])


@dbtest
def test_bools(executor):
    run(executor, """create table test(a boolean)""")
    run(executor, """insert into test values(True)""")
    results = run(executor, """select * from test""")

    assert_result_equal(results, header=["a"], rows=[(1,)])


@dbtest
def test_binary(executor):
    run(executor, """create table bt(geom linestring NOT NULL)""")
    run(executor, "INSERT INTO bt VALUES (ST_GeomFromText('LINESTRING(116.37604 39.73979,116.375 39.73965)'));")
    results = run(executor, """select * from bt""")

    geom = (
        b"\x00\x00\x00\x00\x01\x02\x00\x00\x00\x02\x00\x00\x009\x7f\x13\n"
        b"\x11\x18]@4\xf4Op\xb1\xdeC@\x00\x00\x00\x00\x00\x18]@B>\xe8\xd9"
        b"\xac\xdeC@"
    )

    assert_result_equal(results, header=["geom"], rows=[(geom,)])


@dbtest
def test_table_and_columns_query(executor):
    run(executor, "create table a(x text, y text)")
    run(executor, "create table b(z text)")

    assert set(executor.tables()) == {("a",), ("b",)}
    assert set(executor.table_columns()) == {("a", "x"), ("a", "y"), ("b", "z")}


@dbtest
def test_database_list(executor):
    databases = executor.databases()
    assert TEST_DATABASE in databases


@dbtest
def test_invalid_syntax(executor):
    with pytest.raises(pymysql.ProgrammingError) as excinfo:
        run(executor, "invalid syntax!")
    assert "You have an error in your SQL syntax;" in str(excinfo.value)


@dbtest
def test_invalid_column_name(executor):
    with pytest.raises(pymysql.err.OperationalError) as excinfo:
        run(executor, "select invalid command")
    assert "Unknown column 'invalid' in 'field list'" in str(excinfo.value)


@dbtest
def test_unicode_support_in_output(executor):
    run(executor, "create table unicodechars(t text)")
    run(executor, "insert into unicodechars (t) values ('é')")

    # See issue #24, this raises an exception without proper handling
    results = run(executor, "select * from unicodechars")
    assert_result_equal(results, header=["t"], rows=[("é",)])


@dbtest
def test_multiple_queries_same_line(executor):
    results = run(executor, "select 'foo'; select 'bar'")

    expected = [
        {
            "preamble": None,
            "header": ["foo"],
            "rows": [("foo",)],
            "postamble": None,
            "status_plain": "1 row in set",
            'status': FormattedText([('', '1 row in set')]),
        },
        {
            "preamble": None,
            "header": ["bar"],
            "rows": [("bar",)],
            "postamble": None,
            "status_plain": "1 row in set",
            'status': FormattedText([('', '1 row in set')]),
        },
    ]
    assert expected == results


@dbtest
def test_multiple_queries_same_line_syntaxerror(executor):
    with pytest.raises(pymysql.ProgrammingError) as excinfo:
        run(executor, "select 'foo'; invalid syntax")
    assert "You have an error in your SQL syntax;" in str(excinfo.value)


@dbtest
@pytest.mark.skipif(os.name == "nt", reason="Bug: fails on Windows, needs fixing, singleton of FQ not working right")
def test_favorite_query(executor, monkeypatch):
    monkeypatch.setattr(iocommands.FavoriteQueries, 'instance', iocommands.favoritequeries, raising=False)
    set_expanded_output(False)
    run(executor, "create table test(a text)")
    run(executor, "insert into test values('abc')")
    run(executor, "insert into test values('def')")

    results = run(executor, "\\fs test-a select * from test where a like 'a%'")
    assert_result_equal(results, status="Saved.", status_plain="Saved.")

    results = run(executor, "\\f test-a")
    assert_result_equal(results, preamble="> select * from test where a like 'a%'", header=["a"], rows=[("abc",)], auto_status=False)

    results = run(executor, "\\fd test-a")
    assert_result_equal(results, status="test-a: Deleted.", status_plain="test-a: Deleted.")


@dbtest
@pytest.mark.skipif(os.name == "nt", reason="Bug: fails on Windows, needs fixing, singleton of FQ not working right")
def test_favorite_query_multiple_statement(executor, monkeypatch):
    monkeypatch.setattr(iocommands.FavoriteQueries, 'instance', iocommands.favoritequeries, raising=False)
    set_expanded_output(False)
    run(executor, "create table test(a text)")
    run(executor, "insert into test values('abc')")
    run(executor, "insert into test values('def')")

    results = run(executor, "\\fs test-ad select * from test where a like 'a%'; select * from test where a like 'd%'")
    assert_result_equal(results, status="Saved.", status_plain="Saved.")

    results = run(executor, "\\f test-ad")
    expected = [
        {
            "preamble": "> select * from test where a like 'a%'",
            "header": ["a"],
            "rows": [("abc",)],
            "postamble": None,
            "status": None,
            "status_plain": None,
        },
        {
            "preamble": "> select * from test where a like 'd%'",
            "header": ["a"],
            "rows": [("def",)],
            "postamble": None,
            "status": None,
            "status_plain": None,
        },
    ]
    assert expected == results

    results = run(executor, "\\fd test-ad")
    assert_result_equal(results, status="test-ad: Deleted.", status_plain="test-ad: Deleted.")


@dbtest
@pytest.mark.skipif(os.name == "nt", reason="Bug: fails on Windows, needs fixing, singleton of FQ not working right")
def test_favorite_query_expanded_output(executor, monkeypatch):
    monkeypatch.setattr(iocommands.FavoriteQueries, 'instance', iocommands.favoritequeries, raising=False)
    set_expanded_output(False)
    run(executor, """create table test(a text)""")
    run(executor, """insert into test values('abc')""")

    results = run(executor, "\\fs test-ae select * from test")
    assert_result_equal(results, status="Saved.", status_plain="Saved.")

    results = run(executor, "\\f test-ae \\G")
    assert is_expanded_output() is True
    assert_result_equal(results, preamble="> select * from test", header=["a"], rows=[("abc",)], auto_status=False)

    set_expanded_output(False)

    results = run(executor, "\\fd test-ae")
    assert_result_equal(results, status="test-ae: Deleted.", status_plain="test-ae: Deleted.")


@dbtest
def test_collapsed_output_special_command(executor):
    set_expanded_output(True)
    run(executor, "select 1\\g")
    assert is_expanded_output() is False


@dbtest
def test_special_command(executor):
    results = run(executor, "\\?")
    assert_result_equal(results, rows=("quit", "\\q", "quit", "Quit."), header="Command", assert_contains=True, auto_status=False)


@dbtest
def test_cd_command_without_a_folder_name(executor):
    results = run(executor, "system cd")
    assert_result_equal(
        results, status="Exactly one directory name must be provided.", status_plain="Exactly one directory name must be provided."
    )


@dbtest
@pytest.mark.skipif(os.name == 'nt', reason='todo: unknown')
def test_cd_command_with_one_nonexistent_folder_name(executor):
    results = run(executor, 'system cd nonexistent_folder_name')
    assert_result_equal(results, status='No such file or directory', status_plain='No such file or directory')


@dbtest
def test_cd_command_with_one_real_folder_name(executor, tmp_path, monkeypatch):
    monkeypatch.chdir(tmp_path)
    doc_dir = tmp_path / 'doc'
    doc_dir.mkdir()
    results = run(executor, 'system cd doc')
    # todo would be better to capture stderr but there was a problem with capsys
    assert results[0]['status_plain'] is None


@dbtest
def test_cd_command_with_two_folder_names(executor):
    results = run(executor, "system cd one two")
    assert_result_equal(
        results, status='Exactly one directory name must be provided.', status_plain='Exactly one directory name must be provided.'
    )


@dbtest
def test_cd_command_unbalanced(executor):
    results = run(executor, "system cd 'one")
    assert_result_equal(
        results,
        status='Cannot parse system command: No closing quotation',
        status_plain='Cannot parse system command: No closing quotation',
    )


@dbtest
def test_system_command_not_found(executor):
    results = run(executor, "system xyz")
    if os.name == "nt":
        assert_result_equal(results, status_plain="OSError: The system cannot find the file specified", assert_contains=True)
    else:
        assert_result_equal(results, status_plain="OSError: No such file or directory", assert_contains=True)


@dbtest
def test_system_command_output(executor):
    eol = os.linesep
    results = run(executor, "system echo mycli rocks")
    assert_result_equal(results, preamble=f"mycli rocks{eol}")


@dbtest
def test_cd_command_current_dir(executor):
    test_path = os.path.abspath(os.path.dirname(__file__))
    run(executor, f"system cd {test_path}")
    assert os.getcwd() == test_path


@dbtest
def test_unicode_support(executor):
    results = run(executor, "SELECT '日本語' AS japanese;")
    assert_result_equal(results, header=["japanese"], rows=[("日本語",)])


@dbtest
def test_timestamp_null(executor):
    run(executor, """create table ts_null(a timestamp null)""")
    run(executor, """insert into ts_null values(null)""")
    results = run(executor, """select * from ts_null""")
    assert_result_equal(results, header=["a"], rows=[(None,)])


@dbtest
def test_datetime_null(executor):
    run(executor, """create table dt_null(a datetime null)""")
    run(executor, """insert into dt_null values(null)""")
    results = run(executor, """select * from dt_null""")
    assert_result_equal(results, header=["a"], rows=[(None,)])


@dbtest
def test_date_null(executor):
    run(executor, """create table date_null(a date null)""")
    run(executor, """insert into date_null values(null)""")
    results = run(executor, """select * from date_null""")
    assert_result_equal(results, header=["a"], rows=[(None,)])


@dbtest
def test_time_null(executor):
    run(executor, """create table time_null(a time null)""")
    run(executor, """insert into time_null values(null)""")
    results = run(executor, """select * from time_null""")
    assert_result_equal(results, header=["a"], rows=[(None,)])


@dbtest
def test_multiple_results(executor):
    query = """CREATE PROCEDURE dmtest()
        BEGIN
          SELECT 1;
          SELECT 2;
        END"""
    executor.conn.cursor().execute(query)

    results = run(executor, "call dmtest;")
    expected = [
        {
            "preamble": None,
            "header": ["1"],
            "rows": [(1,)],
            "postamble": None,
            "status_plain": "1 row in set",
            'status': FormattedText([('', '1 row in set')]),
        },
        {
            "preamble": None,
            "header": ["2"],
            "rows": [(2,)],
            "postamble": None,
            "status_plain": "1 row in set",
            'status': FormattedText([('', '1 row in set')]),
        },
    ]
    assert results == expected


@pytest.mark.parametrize(
    "version_string, species, parsed_version_string, version",
    (
        ("5.7.25-TiDB-v6.1.0", "TiDB", "6.1.0", 60100),
        ("8.0.11-TiDB-v7.2.0-alpha-69-g96e9e68daa", "TiDB", "7.2.0", 70200),
        ("5.7.32-35", "Percona", "5.7.32", 50732),
        ("5.7.32-0ubuntu0.18.04.1", "MySQL", "5.7.32", 50732),
        ("10.5.8-MariaDB-1:10.5.8+maria~focal", "MariaDB", "10.5.8", 100508),
        ("5.5.5-10.5.8-MariaDB-1:10.5.8+maria~focal", "MariaDB", "10.5.8", 100508),
        ("5.0.16-pro-nt-log", "MySQL", "5.0.16", 50016),
        ("5.1.5a-alpha", "MySQL", "5.1.5", 50105),
        # Plain X.Y.Z with no suffix (e.g. Homebrew MySQL)
        ("5.7.99", "MySQL", "5.7.99", 50799),
        ("unexpected version string", None, "", 0),
        ("", None, "", 0),
        (None, None, "", 0),
    ),
)
def test_version_parsing(version_string, species, parsed_version_string, version):
    server_info = ServerInfo.from_version_string(version_string)
    assert (server_info.species and server_info.species.name) == species or ServerSpecies.MySQL
    assert server_info.version_str == parsed_version_string
    assert server_info.version == version


@pytest.mark.parametrize(
    'version_string, expected',
    (
        ('5.7.32', 50732),
        ('8.0.11', 80011),
        ('10.5.8', 100508),
    ),
)
def test_calc_mysql_version_value(version_string: str, expected: int) -> None:
    assert ServerInfo.calc_mysql_version_value(version_string) == expected


@pytest.mark.parametrize(
    'version_string',
    (
        None,
        '',
        123,
        '8.0',
        '8.0.11.1',
        'unexpected version string',
    ),
)
def test_calc_mysql_version_value_returns_zero_for_invalid_input(version_string: object) -> None:
    assert ServerInfo.calc_mysql_version_value(version_string) == 0


@pytest.mark.parametrize('version_string', ('8.0.x', '8.x.11', 'x.0.11'))
def test_calc_mysql_version_value_raises_for_non_numeric_parts(version_string: str) -> None:
    with pytest.raises(ValueError):
        ServerInfo.calc_mysql_version_value(version_string)


def test_sqlexecute_import_swallows_optional_dependency_import_errors(monkeypatch) -> None:
    assert sqlexecute.__file__ is not None
    original_import = builtins.__import__

    def fake_import(name, globals=None, locals=None, fromlist=(), level=0):  # noqa: A002
        if name == 'paramiko':
            raise ImportError('missing optional dependency')
        return original_import(name, globals, locals, fromlist, level)

    module_name = 'sqlexecute_importerror_test'
    spec = importlib.util.spec_from_file_location(module_name, Path(sqlexecute.__file__))
    assert spec is not None
    assert spec.loader is not None
    module = importlib.util.module_from_spec(spec)
    monkeypatch.setattr(builtins, '__import__', fake_import)
    sys.modules[module_name] = module
    try:
        spec.loader.exec_module(module)
    finally:
        sys.modules.pop(module_name, None)


@pytest.mark.parametrize(
    ('server_info', 'expected'),
    (
        (ServerInfo(ServerSpecies.MySQL, '8.0.36'), 'MySQL 8.0.36'),
        (ServerInfo(None, '8.0.36'), '8.0.36'),
    ),
)
def test_server_info_string_representation(server_info: ServerInfo, expected: str) -> None:
    assert str(server_info) == expected


@pytest.mark.parametrize(
    'column_type, expected',
    (
        ("enum('small','medium','large')", ["small", "medium", "large"]),
        ("ENUM('yes','no')", ["yes", "no"]),
        ("enum('a,b','c')", ["a,b", "c"]),
        ("enum('it''s','can\\\\t')", ["it's", "can\\t"]),
    ),
)
def test_parse_enum_values(column_type: str, expected: list[str]) -> None:
    assert SQLExecute._parse_enum_values(column_type) == expected


@pytest.mark.parametrize('column_type', ('', 'varchar(255)', "set('a','b')", None))
def test_parse_enum_values_returns_empty_list_for_non_enum_input(column_type: str | None) -> None:
    assert SQLExecute._parse_enum_values(column_type) == []


class DummyConnection:
    def __init__(self, server_version: str, close_error: Exception | None = None) -> None:
        self.server_version = server_version
        self.host = 'initial-host'
        self.port = 3306
        self.close_calls = 0
        self.connect_calls = 0
        self.close_error = close_error

    def close(self) -> None:
        self.close_calls += 1
        if self.close_error is not None:
            raise self.close_error

    def connect(self) -> None:
        self.connect_calls += 1


class FakeQueryCursor:
    def __init__(
        self,
        nextset_steps: list[tuple[bool, int, object | None]] | None = None,
    ) -> None:
        self.executed: list[str] = []
        self.rowcount = 1
        self.description: object | None = [('column',)]
        self.warning_count = 0
        self._nextset_steps = list(nextset_steps or [])

    def execute(self, sql: str) -> None:
        self.executed.append(sql)

    def nextset(self) -> bool:
        if not self._nextset_steps:
            return False

        has_next, rowcount, description = self._nextset_steps.pop(0)
        self.rowcount = rowcount
        self.description = description
        return has_next


class FakeQueryConnection:
    def __init__(self, cursors: list[FakeQueryCursor]) -> None:
        self.cursors = list(cursors)
        self.cursor_calls = 0

    def cursor(self) -> FakeQueryCursor:
        cursor = self.cursors[self.cursor_calls]
        self.cursor_calls += 1
        return cursor


class FakeMetadataCursor:
    def __init__(
        self,
        rows: list[tuple[object, ...]],
        execute_error: Exception | None = None,
    ) -> None:
        self.rows = rows
        self.execute_error = execute_error
        self.executed: list[tuple[str, tuple[object, ...] | None]] = []
        self.entered = False
        self.exited = False

    def __enter__(self) -> 'FakeMetadataCursor':
        self.entered = True
        return self

    def __exit__(self, exc_type: object, exc: object, tb: object) -> None:
        self.exited = True

    def execute(self, sql: str, params: tuple[object, ...] | None = None) -> None:
        self.executed.append((sql, params))
        if self.execute_error is not None:
            raise self.execute_error

    def fetchall(self) -> list[tuple[object, ...]]:
        return self.rows

    def fetchone(self) -> tuple[object, ...] | None:
        if self.rows:
            return self.rows[0]
        return None

    def __iter__(self):
        return iter(self.rows)


class FakeMetadataConnection:
    def __init__(self, cursor: FakeMetadataCursor) -> None:
        self._cursor = cursor

    def cursor(self) -> FakeMetadataCursor:
        return self._cursor


class FakeConnectionIdCursor:
    def __init__(self, row: tuple[int] | None) -> None:
        self.row = row

    def fetchone(self) -> tuple[int] | None:
        return self.row


class FakeSelectableConnection:
    def __init__(self) -> None:
        self.selected_databases: list[str] = []

    def select_db(self, db: str) -> None:
        self.selected_databases.append(db)


class FakeSSLContext:
    def __init__(self) -> None:
        self.check_hostname = True
        self.verify_mode = None
        self.minimum_version = None
        self.maximum_version = None
        self.loaded_cert_chain: tuple[str, str | None] | None = None
        self.cipher_string: str | None = None

    def load_cert_chain(self, certfile: str, keyfile: str | None = None) -> None:
        self.loaded_cert_chain = (certfile, keyfile)

    def set_ciphers(self, cipher_string: str) -> None:
        self.cipher_string = cipher_string


def make_executor_for_connect_tests() -> SQLExecute:
    executor = SQLExecute.__new__(SQLExecute)
    executor.dbname = 'stored_db'
    executor.user = 'stored_user'
    executor.password = 'stored_password'
    executor.host = 'stored_host'
    executor.port = 3306
    executor.socket = '/tmp/mysql.sock'
    executor.character_set = 'utf8mb4'
    executor.local_infile = True
    executor.ssl = {'ca': '/stored/ca.pem'}
    executor.server_info = None
    executor.connection_id = None
    executor.ssh_user = 'stored_ssh_user'
    executor.ssh_host = None
    executor.ssh_port = 22
    executor.ssh_password = 'stored_ssh_password'
    executor.ssh_key_filename = '/stored/key.pem'
    executor.init_command = 'select 1'
    executor.unbuffered = False
    executor.sandbox_mode = False
    executor.conn = None
    return executor


def make_executor_for_run_tests(conn: object | None = None) -> SQLExecute:
    executor = SQLExecute.__new__(SQLExecute)
    executor.conn = conn
    return executor


def test_connect_updates_connection_state_and_merges_overrides(monkeypatch) -> None:
    executor = make_executor_for_connect_tests()
    previous_conn = DummyConnection(
        server_version='5.7.0',
        close_error=pymysql.err.Error(),
    )
    executor.conn = previous_conn

    new_conn = DummyConnection(server_version='8.0.36-0ubuntu0.22.04.1')
    connect_kwargs = {}
    reset_calls = []
    ssl_context = object()
    ssl_params = {'ca': '/override/ca.pem'}

    def fake_connect(**kwargs):
        connect_kwargs.update(kwargs)
        return new_conn

    def fake_create_ssl_ctx(self, sslp):
        assert self is executor
        assert sslp == ssl_params
        return ssl_context

    def fake_reset_connection_id(self) -> None:
        assert self is executor
        reset_calls.append(True)
        self.connection_id = 42

    monkeypatch.setattr(sqlexecute.pymysql, 'connect', fake_connect)
    monkeypatch.setattr(SQLExecute, '_create_ssl_ctx', fake_create_ssl_ctx)
    monkeypatch.setattr(SQLExecute, 'reset_connection_id', fake_reset_connection_id)

    executor.connect(
        database='override_db',
        user='override_user',
        password='override_password',
        host='override_host',
        port=3307,
        character_set='latin1',
        local_infile=False,
        ssl=ssl_params,
        init_command='select 1; select 2',
        unbuffered=True,
    )

    assert connect_kwargs['database'] == 'override_db'
    assert connect_kwargs['user'] == 'override_user'
    assert connect_kwargs['password'] == 'override_password'
    assert connect_kwargs['host'] == 'override_host'
    assert connect_kwargs['port'] == 3307
    assert connect_kwargs['unix_socket'] == '/tmp/mysql.sock'
    assert connect_kwargs['charset'] == 'latin1'
    assert connect_kwargs['local_infile'] is False
    assert connect_kwargs['ssl'] is ssl_context
    assert connect_kwargs['defer_connect'] is False
    assert connect_kwargs['init_command'] == 'select 1; select 2'
    assert connect_kwargs['cursorclass'] is sqlexecute.pymysql.cursors.SSCursor
    assert connect_kwargs['client_flag'] & sqlexecute.pymysql.constants.CLIENT.INTERACTIVE
    assert connect_kwargs['client_flag'] & sqlexecute.pymysql.constants.CLIENT.MULTI_STATEMENTS
    assert connect_kwargs['program_name'] == 'mycli'
    assert previous_conn.close_calls == 1
    assert executor.conn is new_conn
    assert executor.dbname == 'override_db'
    assert executor.user == 'override_user'
    assert executor.password == 'override_password'
    assert executor.host == 'override_host'
    assert executor.port == 3307
    assert executor.socket == '/tmp/mysql.sock'
    assert executor.character_set == 'latin1'
    assert executor.ssl == ssl_params
    assert executor.init_command == 'select 1; select 2'
    assert executor.unbuffered is True
    assert executor.connection_id == 42
    assert reset_calls == [True]
    assert executor.server_info is not None
    assert executor.server_info.version_str == '8.0.36'
    assert executor.server_info.version == 80036


def test_connect_sets_expired_password_flag(monkeypatch) -> None:
    executor = make_executor_for_connect_tests()
    executor.ssl = None

    new_conn = DummyConnection(server_version='8.0.36-0ubuntu0.22.04.1')
    connect_kwargs = {}

    def fake_connect(**kwargs):
        connect_kwargs.update(kwargs)
        return new_conn

    monkeypatch.setattr(sqlexecute.pymysql, 'connect', fake_connect)
    monkeypatch.setattr(SQLExecute, 'reset_connection_id', lambda self: None)

    executor.connect()

    assert connect_kwargs['client_flag'] & sqlexecute.pymysql.constants.CLIENT.HANDLE_EXPIRED_PASSWORDS
    assert executor.sandbox_mode is False


def test_connect_falls_back_to_sandbox_on_1820(monkeypatch) -> None:
    executor = make_executor_for_connect_tests()
    executor.ssl = None

    new_conn = DummyConnection(server_version='8.0.36-0ubuntu0.22.04.1')
    call_count = 0
    sandbox_calls = []

    def fake_connect(**kwargs):
        nonlocal call_count
        call_count += 1
        if call_count == 1:
            raise pymysql.OperationalError(1820, 'must change password')
        return new_conn

    def fake_connect_sandbox(self, conn):
        sandbox_calls.append(conn)

    monkeypatch.setattr(sqlexecute.pymysql, 'connect', fake_connect)
    monkeypatch.setattr(SQLExecute, '_connect_sandbox', fake_connect_sandbox)

    executor.connect()

    assert call_count == 2
    assert len(sandbox_calls) == 1
    assert executor.sandbox_mode is True
    assert executor.server_info is None
    assert executor.connection_id is None


def test_connect_reraises_non_sandbox_operational_error(monkeypatch) -> None:
    executor = make_executor_for_connect_tests()
    executor.ssl = None

    def fake_connect(**_kwargs):
        raise pymysql.OperationalError(1045, 'access denied')

    monkeypatch.setattr(sqlexecute.pymysql, 'connect', fake_connect)

    with pytest.raises(pymysql.OperationalError) as exc_info:
        executor.connect()

    assert exc_info.value.args == (1045, 'access denied')


def test_connect_uses_ssh_tunnel_when_ssh_host_is_set(monkeypatch) -> None:
    executor = make_executor_for_connect_tests()
    executor.ssl = None
    new_conn = DummyConnection(server_version='8.0.36-0ubuntu0.22.04.1')
    connect_kwargs = {}
    tunnel_args = {}
    tunnel_started = []

    class FakeTunnel:
        def __init__(
            self,
            ssh_address_or_host,
            ssh_username=None,
            ssh_pkey=None,
            ssh_password=None,
            remote_bind_address=None,
        ) -> None:
            tunnel_args['ssh_address_or_host'] = ssh_address_or_host
            tunnel_args['ssh_username'] = ssh_username
            tunnel_args['ssh_pkey'] = ssh_pkey
            tunnel_args['ssh_password'] = ssh_password
            tunnel_args['remote_bind_address'] = remote_bind_address
            self.local_bind_host = '127.0.0.1'
            self.local_bind_port = 4406

        def start(self) -> None:
            tunnel_started.append(True)

    def fake_connect(**kwargs):
        connect_kwargs.update(kwargs)
        return new_conn

    def fake_reset_connection_id(self) -> None:
        self.connection_id = 7

    monkeypatch.setattr(sqlexecute.pymysql, 'connect', fake_connect)
    monkeypatch.setattr(SQLExecute, 'reset_connection_id', fake_reset_connection_id)
    monkeypatch.setattr(
        sqlexecute,
        'sshtunnel',
        SimpleNamespace(SSHTunnelForwarder=FakeTunnel),
        raising=False,
    )

    executor.connect(
        host='db.internal',
        port=3308,
        ssh_host='bastion.internal',
        ssh_port=2222,
        ssh_user='alice',
        ssh_password='secret',
        ssh_key_filename='/tmp/id_rsa',
    )

    assert connect_kwargs['host'] == 'db.internal'
    assert connect_kwargs['port'] == 3308
    assert connect_kwargs['defer_connect'] is True
    assert connect_kwargs['init_command'] == 'select 1'
    assert tunnel_args['ssh_address_or_host'] == ('bastion.internal', 2222)
    assert tunnel_args['ssh_username'] == 'alice'
    assert tunnel_args['ssh_pkey'] == '/tmp/id_rsa'
    assert tunnel_args['ssh_password'] == 'secret'
    assert tunnel_args['remote_bind_address'] == ('db.internal', 3308)
    assert tunnel_started == [True]
    assert new_conn.host == '127.0.0.1'
    assert new_conn.port == 4406
    assert new_conn.connect_calls == 1
    assert executor.conn is new_conn
    assert executor.host == 'db.internal'
    assert executor.port == 3308
    assert executor.connection_id == 7


def test_connect_reraises_ssh_tunnel_errors(monkeypatch) -> None:
    executor = make_executor_for_connect_tests()
    executor.ssl = None
    new_conn = DummyConnection(server_version='8.0.36-0ubuntu0.22.04.1')

    class FakeTunnel:
        def __init__(self, *args, **kwargs) -> None:
            self.local_bind_host = '127.0.0.1'
            self.local_bind_port = 4406

        def start(self) -> None:
            raise RuntimeError('tunnel failed')

    monkeypatch.setattr(sqlexecute.pymysql, 'connect', lambda **_kwargs: new_conn)
    monkeypatch.setattr(
        sqlexecute,
        'sshtunnel',
        SimpleNamespace(SSHTunnelForwarder=FakeTunnel),
        raising=False,
    )

    with pytest.raises(RuntimeError, match='tunnel failed'):
        executor.connect(ssh_host='bastion.internal')


def test_connect_sandbox_temporarily_disables_set_character_set() -> None:
    original_calls = []
    connect_observed_stub = []

    class FakeSandboxConnection:
        def set_character_set(self, *args, **kwargs) -> None:
            original_calls.append((args, kwargs))

        def connect(self) -> None:
            self.set_character_set('utf8mb4')
            connect_observed_stub.append(original_calls == [])

    conn = FakeSandboxConnection()
    original_set_character_set = conn.set_character_set

    SQLExecute._connect_sandbox(conn)

    assert connect_observed_stub == [True]
    assert conn.set_character_set == original_set_character_set
    conn.set_character_set('latin1')
    assert original_calls == [(('latin1',), {})]


def test_run_returns_empty_result_for_blank_statement(monkeypatch) -> None:
    split_inputs: list[str] = []

    def fake_split_queries(statement: str):
        split_inputs.append(statement)
        return iter(())

    monkeypatch.setattr(sqlexecute.iocommands, 'split_queries', fake_split_queries)

    executor = make_executor_for_run_tests()

    assert list(executor.run('   \n\t  ')) == [SQLResult()]
    assert split_inputs == ['']


def test_run_does_not_split_favorite_query(monkeypatch) -> None:
    favorite_results = [SQLResult(status='Saved.')]
    favorite_sql = '\\fs test-name select 1; select 2'
    cursor = FakeQueryCursor()
    execute_calls: list[str] = []

    def fake_execute(cur: FakeQueryCursor, sql: str) -> list[SQLResult]:
        assert cur is cursor
        execute_calls.append(sql)
        return favorite_results

    def fail_split_queries(_statement: str):
        raise AssertionError('split_queries() should not be called for favorite queries')

    monkeypatch.setattr(sqlexecute, 'Connection', FakeQueryConnection)
    monkeypatch.setattr(sqlexecute, 'execute', fake_execute)
    monkeypatch.setattr(sqlexecute.iocommands, 'split_queries', fail_split_queries)

    executor = make_executor_for_run_tests(FakeQueryConnection([cursor]))

    assert list(executor.run(favorite_sql)) == favorite_results
    assert execute_calls == [favorite_sql]
    assert cursor.executed == []


def test_run_uses_special_command_results_without_regular_execution(monkeypatch) -> None:
    cursor = FakeQueryCursor()
    special_results = [SQLResult(status='special command')]

    def fake_execute(cur: FakeQueryCursor, sql: str) -> list[SQLResult]:
        assert cur is cursor
        assert sql == '\\dt'
        return special_results

    def fail_get_result(_self: SQLExecute, _cursor: object) -> SQLResult:
        raise AssertionError('get_result() should not be called for handled special commands')

    monkeypatch.setattr(sqlexecute, 'Connection', FakeQueryConnection)
    monkeypatch.setattr(sqlexecute, 'execute', fake_execute)
    monkeypatch.setattr(sqlexecute.iocommands, 'split_queries', lambda statement: iter([statement]))
    monkeypatch.setattr(SQLExecute, 'get_result', fail_get_result)

    executor = make_executor_for_run_tests(FakeQueryConnection([cursor]))

    assert list(executor.run('\\dt')) == special_results
    assert cursor.executed == []


def test_run_falls_back_to_regular_sql_and_handles_output_flags(monkeypatch) -> None:
    cursors = [FakeQueryCursor(), FakeQueryCursor()]
    expanded_values: list[bool] = []
    forced_horizontal_values: list[bool] = []
    get_result_calls: list[list[str]] = []

    def fake_execute(_cur: FakeQueryCursor, _sql: str) -> list[SQLResult]:
        raise sqlexecute.CommandNotFound('not a special command')

    def fake_get_result(_self: SQLExecute, cursor: FakeQueryCursor) -> SQLResult:
        get_result_calls.append(list(cursor.executed))
        return SQLResult(status=f'ran {cursor.executed[-1]}')

    monkeypatch.setattr(sqlexecute, 'Connection', FakeQueryConnection)
    monkeypatch.setattr(sqlexecute, 'execute', fake_execute)
    monkeypatch.setattr(
        sqlexecute.iocommands,
        'split_queries',
        lambda _statement: iter(['select 1\\G', 'select 2\\g']),
    )
    monkeypatch.setattr(
        sqlexecute.iocommands,
        'set_expanded_output',
        lambda value: expanded_values.append(value),
    )
    monkeypatch.setattr(
        sqlexecute.iocommands,
        'set_forced_horizontal_output',
        lambda value: forced_horizontal_values.append(value),
    )
    monkeypatch.setattr(SQLExecute, 'get_result', fake_get_result)

    executor = make_executor_for_run_tests(FakeQueryConnection(cursors))

    results = list(executor.run('select 1; select 2'))

    assert [result.status for result in results] == ['ran select 1', 'ran select 2']
    assert expanded_values == [True, False]
    assert forced_horizontal_values == [True]
    assert [cursor.executed for cursor in cursors] == [['select 1'], ['select 2']]
    assert get_result_calls == [['select 1'], ['select 2']]


def test_run_yields_each_non_empty_result_set_until_nextset_is_false(monkeypatch) -> None:
    cursor = FakeQueryCursor(
        nextset_steps=[
            (True, 1, [('column',)]),
            (False, 1, [('column',)]),
        ]
    )
    get_result_calls: list[int] = []

    def fake_execute(_cur: FakeQueryCursor, _sql: str) -> list[SQLResult]:
        raise sqlexecute.CommandNotFound('not a special command')

    def fake_get_result(_self: SQLExecute, _cursor: FakeQueryCursor) -> SQLResult:
        get_result_calls.append(len(get_result_calls) + 1)
        return SQLResult(status=f'result {len(get_result_calls)}')

    monkeypatch.setattr(sqlexecute, 'Connection', FakeQueryConnection)
    monkeypatch.setattr(sqlexecute, 'execute', fake_execute)
    monkeypatch.setattr(sqlexecute.iocommands, 'split_queries', lambda statement: iter([statement]))
    monkeypatch.setattr(SQLExecute, 'get_result', fake_get_result)

    executor = make_executor_for_run_tests(FakeQueryConnection([cursor]))

    results = list(executor.run('call demo()'))

    assert [result.status for result in results] == ['result 1', 'result 2']
    assert cursor.executed == ['call demo()']
    assert get_result_calls == [1, 2]


def test_run_skips_trailing_empty_result_set_from_nextset(monkeypatch) -> None:
    cursor = FakeQueryCursor(nextset_steps=[(True, 0, None)])
    get_result_calls: list[int] = []

    def fake_execute(_cur: FakeQueryCursor, _sql: str) -> list[SQLResult]:
        raise sqlexecute.CommandNotFound('not a special command')

    def fake_get_result(_self: SQLExecute, _cursor: FakeQueryCursor) -> SQLResult:
        get_result_calls.append(1)
        return SQLResult(status='result 1')

    monkeypatch.setattr(sqlexecute, 'Connection', FakeQueryConnection)
    monkeypatch.setattr(sqlexecute, 'execute', fake_execute)
    monkeypatch.setattr(sqlexecute.iocommands, 'split_queries', lambda statement: iter([statement]))
    monkeypatch.setattr(SQLExecute, 'get_result', fake_get_result)

    executor = make_executor_for_run_tests(FakeQueryConnection([cursor]))

    results = list(executor.run('call demo()'))

    assert [result.status for result in results] == ['result 1']
    assert cursor.executed == ['call demo()']
    assert get_result_calls == [1]


def test_get_result_returns_header_and_row_status_for_result_sets() -> None:
    cursor = FakeQueryCursor()
    cursor.rowcount = 2
    cursor.description = [('name',), ('age',)]
    cursor.warning_count = 0

    executor = make_executor_for_run_tests()

    result = executor.get_result(cursor)

    assert result.preamble is None
    assert result.header == ['name', 'age']
    assert result.rows is cursor
    assert result.postamble is None
    assert result.status_plain == '2 rows in set'


def test_get_result_returns_query_ok_status_when_no_result_set() -> None:
    cursor = FakeQueryCursor()
    cursor.rowcount = 1
    cursor.description = None
    cursor.warning_count = 0

    executor = make_executor_for_run_tests()

    result = executor.get_result(cursor)

    assert result.header is None
    assert result.rows is cursor
    assert result.status_plain == 'Query OK, 1 row affected'


def test_get_result_appends_warning_count_to_status() -> None:
    cursor = FakeQueryCursor()
    cursor.rowcount = 3
    cursor.description = [('name',)]
    cursor.warning_count = 2

    executor = make_executor_for_run_tests()

    result = executor.get_result(cursor)

    assert result.header == ['name']
    assert result.rows is cursor
    assert result.status_plain == '3 rows in set, 2 warnings'


def test_tables_executes_show_tables_query_and_yields_rows(monkeypatch) -> None:
    cursor = FakeMetadataCursor([('users',), ('orders',)])
    executor = make_executor_for_run_tests(FakeMetadataConnection(cursor))
    monkeypatch.setattr(sqlexecute, 'Connection', FakeMetadataConnection)

    result = list(executor.tables())

    assert result == [('users',), ('orders',)]
    assert cursor.executed == [(SQLExecute.tables_query, None)]
    assert cursor.entered is True
    assert cursor.exited is True


def test_tables_returns_empty_generator_when_no_tables_exist(monkeypatch) -> None:
    cursor = FakeMetadataCursor([])
    executor = make_executor_for_run_tests(FakeMetadataConnection(cursor))
    monkeypatch.setattr(sqlexecute, 'Connection', FakeMetadataConnection)

    result = list(executor.tables())

    assert result == []
    assert cursor.executed == [(SQLExecute.tables_query, None)]


def test_table_columns_executes_query_with_dbname_and_yields_rows(monkeypatch) -> None:
    cursor = FakeMetadataCursor([('users', 'id'), ('users', 'email'), ('orders', 'id')])
    executor = make_executor_for_run_tests(FakeMetadataConnection(cursor))
    executor.dbname = 'app_db'
    monkeypatch.setattr(sqlexecute, 'Connection', FakeMetadataConnection)

    result = list(executor.table_columns())

    assert result == [('users', 'id'), ('users', 'email'), ('orders', 'id')]
    assert cursor.executed == [(SQLExecute.table_columns_query, ('app_db',))]
    assert cursor.entered is True
    assert cursor.exited is True


def test_table_columns_returns_empty_generator_when_schema_has_no_tables(monkeypatch) -> None:
    cursor = FakeMetadataCursor([])
    executor = make_executor_for_run_tests(FakeMetadataConnection(cursor))
    executor.dbname = 'empty_db'
    monkeypatch.setattr(sqlexecute, 'Connection', FakeMetadataConnection)

    result = list(executor.table_columns())

    assert result == []
    assert cursor.executed == [(SQLExecute.table_columns_query, ('empty_db',))]


def test_enum_values_executes_query_and_skips_non_enum_columns(monkeypatch) -> None:
    cursor = FakeMetadataCursor([
        ('orders', 'status', "enum('new','paid')"),
        ('orders', 'notes', 'varchar(255)'),
    ])
    executor = make_executor_for_run_tests(FakeMetadataConnection(cursor))
    executor.dbname = 'app_db'
    monkeypatch.setattr(sqlexecute, 'Connection', FakeMetadataConnection)

    result = list(executor.enum_values())

    assert result == [('orders', 'status', ['new', 'paid'])]
    assert cursor.executed == [(SQLExecute.enum_values_query, ('app_db',))]
    assert cursor.entered is True
    assert cursor.exited is True


def test_enum_values_returns_empty_generator_when_no_enum_values_are_found(monkeypatch) -> None:
    cursor = FakeMetadataCursor([('orders', 'notes', 'varchar(255)')])
    executor = make_executor_for_run_tests(FakeMetadataConnection(cursor))
    executor.dbname = 'empty_db'
    monkeypatch.setattr(sqlexecute, 'Connection', FakeMetadataConnection)

    result = list(executor.enum_values())

    assert result == []
    assert cursor.executed == [(SQLExecute.enum_values_query, ('empty_db',))]


def test_foreign_keys_executes_query_with_dbname_and_yields_rows(monkeypatch) -> None:
    cursor = FakeMetadataCursor([
        ('orders', 'customer_id', 'customers', 'id'),
        ('order_items', 'order_id', 'orders', 'id'),
    ])
    executor = make_executor_for_run_tests(FakeMetadataConnection(cursor))
    executor.dbname = 'app_db'
    monkeypatch.setattr(sqlexecute, 'Connection', FakeMetadataConnection)

    result = list(executor.foreign_keys())

    assert result == [
        ('orders', 'customer_id', 'customers', 'id'),
        ('order_items', 'order_id', 'orders', 'id'),
    ]
    assert cursor.executed == [(SQLExecute.foreign_keys_query, ('app_db',))]
    assert cursor.entered is True
    assert cursor.exited is True


def test_foreign_keys_returns_empty_generator_and_logs_execute_errors(monkeypatch, caplog) -> None:
    cursor = FakeMetadataCursor([], execute_error=RuntimeError('boom'))
    executor = make_executor_for_run_tests(FakeMetadataConnection(cursor))
    executor.dbname = 'app_db'
    monkeypatch.setattr(sqlexecute, 'Connection', FakeMetadataConnection)

    with caplog.at_level('ERROR', logger='mycli.sqlexecute'):
        result = list(executor.foreign_keys())

    assert result == []
    assert cursor.executed == [(SQLExecute.foreign_keys_query, ('app_db',))]
    assert cursor.entered is True
    assert cursor.exited is True
    assert "No foreign key completions due to RuntimeError('boom')" in caplog.text


def test_databases_executes_show_databases_and_flattens_names(monkeypatch) -> None:
    cursor = FakeMetadataCursor([('mysql',), ('information_schema',), ('app_db',)])
    executor = make_executor_for_run_tests(FakeMetadataConnection(cursor))
    monkeypatch.setattr(sqlexecute, 'Connection', FakeMetadataConnection)

    result = executor.databases()

    assert result == ['mysql', 'information_schema', 'app_db']
    assert cursor.executed == [(SQLExecute.databases_query, None)]
    assert cursor.entered is True
    assert cursor.exited is True


def test_databases_returns_empty_list_when_no_databases_are_found(monkeypatch) -> None:
    cursor = FakeMetadataCursor([])
    executor = make_executor_for_run_tests(FakeMetadataConnection(cursor))
    monkeypatch.setattr(sqlexecute, 'Connection', FakeMetadataConnection)

    result = executor.databases()

    assert result == []
    assert cursor.executed == [(SQLExecute.databases_query, None)]


def test_functions_executes_query_with_dbname_and_yields_rows(monkeypatch) -> None:
    cursor = FakeMetadataCursor([('calculate_total',), ('format_order',)])
    executor = make_executor_for_run_tests(FakeMetadataConnection(cursor))
    executor.dbname = 'app_db'
    monkeypatch.setattr(sqlexecute, 'Connection', FakeMetadataConnection)

    result = list(executor.functions())

    assert result == [('calculate_total',), ('format_order',)]
    assert cursor.executed == [(SQLExecute.functions_query, ('app_db',))]
    assert cursor.entered is True
    assert cursor.exited is True


def test_functions_returns_empty_generator_when_schema_has_no_functions(monkeypatch) -> None:
    cursor = FakeMetadataCursor([])
    executor = make_executor_for_run_tests(FakeMetadataConnection(cursor))
    executor.dbname = 'empty_db'
    monkeypatch.setattr(sqlexecute, 'Connection', FakeMetadataConnection)

    result = list(executor.functions())

    assert result == []
    assert cursor.executed == [(SQLExecute.functions_query, ('empty_db',))]


def test_procedures_executes_query_with_dbname_and_yields_rows(monkeypatch) -> None:
    cursor = FakeMetadataCursor([('refresh_orders',), ('archive_orders',)])
    executor = make_executor_for_run_tests(FakeMetadataConnection(cursor))
    executor.dbname = 'app_db'
    monkeypatch.setattr(sqlexecute, 'Connection', FakeMetadataConnection)

    result = list(executor.procedures())

    assert result == [('refresh_orders',), ('archive_orders',)]
    assert cursor.executed == [(SQLExecute.procedures_query, ('app_db',))]
    assert cursor.entered is True
    assert cursor.exited is True


def test_procedures_yields_empty_tuple_and_logs_database_errors(monkeypatch, caplog) -> None:
    cursor = FakeMetadataCursor([], execute_error=pymysql.DatabaseError('boom'))
    executor = make_executor_for_run_tests(FakeMetadataConnection(cursor))
    executor.dbname = 'app_db'
    monkeypatch.setattr(sqlexecute, 'Connection', FakeMetadataConnection)

    with caplog.at_level('ERROR', logger='mycli.sqlexecute'):
        result = list(executor.procedures())

    assert result == [()]
    assert cursor.executed == [(SQLExecute.procedures_query, ('app_db',))]
    assert cursor.entered is True
    assert cursor.exited is True
    assert "No procedure completions due to DatabaseError('boom')" in caplog.text


def test_character_sets_executes_query_and_yields_rows(monkeypatch) -> None:
    cursor = FakeMetadataCursor([('utf8mb4',), ('latin1',)])
    executor = make_executor_for_run_tests(FakeMetadataConnection(cursor))
    monkeypatch.setattr(sqlexecute, 'Connection', FakeMetadataConnection)

    result = list(executor.character_sets())

    assert result == [('utf8mb4',), ('latin1',)]
    assert cursor.executed == [(SQLExecute.character_sets_query, None)]
    assert cursor.entered is True
    assert cursor.exited is True


def test_character_sets_yields_empty_tuple_and_logs_database_errors(monkeypatch, caplog) -> None:
    cursor = FakeMetadataCursor([], execute_error=pymysql.DatabaseError('boom'))
    executor = make_executor_for_run_tests(FakeMetadataConnection(cursor))
    monkeypatch.setattr(sqlexecute, 'Connection', FakeMetadataConnection)

    with caplog.at_level('ERROR', logger='mycli.sqlexecute'):
        result = list(executor.character_sets())

    assert result == [()]
    assert cursor.executed == [(SQLExecute.character_sets_query, None)]
    assert cursor.entered is True
    assert cursor.exited is True
    assert "No character_set completions due to DatabaseError('boom')" in caplog.text


def test_collations_executes_query_and_yields_rows(monkeypatch) -> None:
    cursor = FakeMetadataCursor([('utf8mb4_general_ci',), ('latin1_swedish_ci',)])
    executor = make_executor_for_run_tests(FakeMetadataConnection(cursor))
    monkeypatch.setattr(sqlexecute, 'Connection', FakeMetadataConnection)

    result = list(executor.collations())

    assert result == [('utf8mb4_general_ci',), ('latin1_swedish_ci',)]
    assert cursor.executed == [(SQLExecute.collations_query, None)]
    assert cursor.entered is True
    assert cursor.exited is True


def test_collations_yields_empty_tuple_and_logs_database_errors(monkeypatch, caplog) -> None:
    cursor = FakeMetadataCursor([], execute_error=pymysql.DatabaseError('boom'))
    executor = make_executor_for_run_tests(FakeMetadataConnection(cursor))
    monkeypatch.setattr(sqlexecute, 'Connection', FakeMetadataConnection)

    with caplog.at_level('ERROR', logger='mycli.sqlexecute'):
        result = list(executor.collations())

    assert result == [()]
    assert cursor.executed == [(SQLExecute.collations_query, None)]
    assert cursor.entered is True
    assert cursor.exited is True
    assert "No collations completions due to DatabaseError('boom')" in caplog.text


def test_show_candidates_executes_query_and_strips_show_prefix(monkeypatch) -> None:
    cursor = FakeMetadataCursor([('SHOW DATABASES',), ('SHOW FULL TABLES',)])
    executor = make_executor_for_run_tests(FakeMetadataConnection(cursor))
    monkeypatch.setattr(sqlexecute, 'Connection', FakeMetadataConnection)

    result = list(executor.show_candidates())

    assert result == [('DATABASES',), ('FULL TABLES',)]
    assert cursor.executed == [(SQLExecute.show_candidates_query, None)]
    assert cursor.entered is True
    assert cursor.exited is True


def test_show_candidates_yields_empty_tuple_and_logs_database_errors(monkeypatch, caplog) -> None:
    cursor = FakeMetadataCursor([], execute_error=pymysql.DatabaseError('boom'))
    executor = make_executor_for_run_tests(FakeMetadataConnection(cursor))
    monkeypatch.setattr(sqlexecute, 'Connection', FakeMetadataConnection)

    with caplog.at_level('ERROR', logger='mycli.sqlexecute'):
        result = list(executor.show_candidates())

    assert result == [()]
    assert cursor.executed == [(SQLExecute.show_candidates_query, None)]
    assert cursor.entered is True
    assert cursor.exited is True
    assert "No show completions due to DatabaseError('boom')" in caplog.text


def test_users_executes_query_and_yields_rows(monkeypatch) -> None:
    cursor = FakeMetadataCursor([("'alice'@'localhost'",), ("'bob'@'%'",)])
    executor = make_executor_for_run_tests(FakeMetadataConnection(cursor))
    monkeypatch.setattr(sqlexecute, 'Connection', FakeMetadataConnection)

    result = list(executor.users())

    assert result == [("'alice'@'localhost'",), ("'bob'@'%'",)]
    assert cursor.executed == [(SQLExecute.users_query, None)]
    assert cursor.entered is True
    assert cursor.exited is True


def test_users_yields_empty_tuple_and_logs_database_errors(monkeypatch, caplog) -> None:
    cursor = FakeMetadataCursor([], execute_error=pymysql.DatabaseError('boom'))
    executor = make_executor_for_run_tests(FakeMetadataConnection(cursor))
    monkeypatch.setattr(sqlexecute, 'Connection', FakeMetadataConnection)

    with caplog.at_level('ERROR', logger='mycli.sqlexecute'):
        result = list(executor.users())

    assert result == [()]
    assert cursor.executed == [(SQLExecute.users_query, None)]
    assert cursor.entered is True
    assert cursor.exited is True
    assert "No user completions due to DatabaseError('boom')" in caplog.text


def test_now_returns_database_timestamp_from_first_row(monkeypatch) -> None:
    timestamp = sqlexecute.datetime.datetime(2024, 1, 2, 3, 4, 5)
    cursor = FakeMetadataCursor([(timestamp,)])
    executor = make_executor_for_run_tests(FakeMetadataConnection(cursor))
    monkeypatch.setattr(sqlexecute, 'Connection', FakeMetadataConnection)

    result = executor.now()

    assert result == timestamp
    assert cursor.executed == [(SQLExecute.now_query, None)]
    assert cursor.entered is True
    assert cursor.exited is True


def test_now_falls_back_to_local_datetime_when_query_returns_no_rows(monkeypatch) -> None:
    fallback = sqlexecute.datetime.datetime(2024, 6, 7, 8, 9, 10)
    cursor = FakeMetadataCursor([])
    executor = make_executor_for_run_tests(FakeMetadataConnection(cursor))

    class FakeDateTime:
        @classmethod
        def now(cls) -> sqlexecute.datetime.datetime:
            return fallback

    monkeypatch.setattr(sqlexecute, 'Connection', FakeMetadataConnection)
    monkeypatch.setattr(sqlexecute.datetime, 'datetime', FakeDateTime)

    result = executor.now()

    assert result == fallback
    assert cursor.executed == [(SQLExecute.now_query, None)]


def test_get_connection_id_returns_cached_value_without_reset(monkeypatch) -> None:
    executor = make_executor_for_run_tests()
    executor.connection_id = 123

    def fail_reset_connection_id(self) -> None:
        raise AssertionError('reset_connection_id() should not be called')

    monkeypatch.setattr(SQLExecute, 'reset_connection_id', fail_reset_connection_id)

    assert executor.get_connection_id() == 123


def test_get_connection_id_resets_when_connection_id_is_missing(monkeypatch) -> None:
    executor = make_executor_for_run_tests()
    executor.connection_id = None
    reset_calls: list[bool] = []

    def fake_reset_connection_id(self) -> None:
        reset_calls.append(True)
        self.connection_id = 456

    monkeypatch.setattr(SQLExecute, 'reset_connection_id', fake_reset_connection_id)

    assert executor.get_connection_id() == 456
    assert reset_calls == [True]


def test_reset_connection_id_sets_connection_id_from_query_result(monkeypatch) -> None:
    executor = make_executor_for_run_tests()
    executor.connection_id = None
    run_calls: list[str] = []

    def fake_run(sql: str):
        run_calls.append(sql)
        return [SimpleNamespace(rows=FakeConnectionIdCursor((789,)))]

    monkeypatch.setattr(sqlexecute, 'Cursor', FakeConnectionIdCursor)
    monkeypatch.setattr(executor, 'run', fake_run)

    executor.reset_connection_id()

    assert executor.connection_id == 789
    assert run_calls == ['select connection_id()']


def test_reset_connection_id_sets_minus_one_when_query_returns_no_row(monkeypatch) -> None:
    executor = make_executor_for_run_tests()
    executor.connection_id = None

    monkeypatch.setattr(sqlexecute, 'Cursor', FakeConnectionIdCursor)
    monkeypatch.setattr(
        executor,
        'run',
        lambda _sql: [SimpleNamespace(rows=FakeConnectionIdCursor(None))],
    )

    executor.reset_connection_id()

    assert executor.connection_id == -1


def test_reset_connection_id_leaves_connection_id_unset_when_query_returns_no_results(monkeypatch) -> None:
    executor = make_executor_for_run_tests()
    executor.connection_id = None

    monkeypatch.setattr(executor, 'run', lambda _sql: iter(()))

    executor.reset_connection_id()

    assert executor.connection_id is None


def test_reset_connection_id_sets_minus_one_and_logs_errors_for_invalid_results(monkeypatch, caplog) -> None:
    executor = make_executor_for_run_tests()
    executor.connection_id = None

    monkeypatch.setattr(sqlexecute, 'Cursor', FakeConnectionIdCursor)
    monkeypatch.setattr(executor, 'run', lambda _sql: [SimpleNamespace(rows=object())])

    with caplog.at_level('ERROR', logger='mycli.sqlexecute'):
        executor.reset_connection_id()

    assert executor.connection_id == -1
    assert 'Failed to get connection id:' in caplog.text


def test_change_db_selects_database_and_updates_dbname(monkeypatch) -> None:
    conn = FakeSelectableConnection()
    executor = make_executor_for_run_tests(conn)
    executor.dbname = 'old_db'
    monkeypatch.setattr(sqlexecute, 'Connection', FakeSelectableConnection)

    executor.change_db('new_db')

    assert conn.selected_databases == ['new_db']
    assert executor.dbname == 'new_db'


def test_create_ssl_ctx_without_ca_disables_hostname_check_and_verification(monkeypatch) -> None:
    executor = make_executor_for_run_tests()
    ctx = FakeSSLContext()
    create_default_context_calls: list[tuple[str | None, str | None]] = []

    def fake_create_default_context(cafile: str | None = None, capath: str | None = None) -> FakeSSLContext:
        create_default_context_calls.append((cafile, capath))
        return ctx

    monkeypatch.setattr(sqlexecute.ssl, 'create_default_context', fake_create_default_context)

    result = executor._create_ssl_ctx({})

    assert result is ctx
    assert create_default_context_calls == [(None, None)]
    assert ctx.check_hostname is False
    assert ctx.verify_mode == sqlexecute.ssl.CERT_NONE
    assert ctx.minimum_version == sqlexecute.ssl.TLSVersion.TLSv1_2
    assert ctx.maximum_version is None
    assert ctx.loaded_cert_chain is None
    assert ctx.cipher_string is None


def test_create_ssl_ctx_applies_cert_cipher_and_tls_version(monkeypatch) -> None:
    executor = make_executor_for_run_tests()
    ctx = FakeSSLContext()
    create_default_context_calls: list[tuple[str | None, str | None]] = []

    def fake_create_default_context(cafile: str | None = None, capath: str | None = None) -> FakeSSLContext:
        create_default_context_calls.append((cafile, capath))
        return ctx

    monkeypatch.setattr(
        sqlexecute.ssl,
        'create_default_context',
        fake_create_default_context,
    )

    result = executor._create_ssl_ctx({
        'ca': '/tmp/ca.pem',
        'check_hostname': False,
        'cert': '/tmp/client-cert.pem',
        'key': '/tmp/client-key.pem',
        'cipher': 'ECDHE-RSA-AES256-GCM-SHA384',
        'tls_version': 'TLSv1.3',
    })

    assert result is ctx
    assert create_default_context_calls == [('/tmp/ca.pem', None)]
    assert ctx.check_hostname is False
    assert ctx.verify_mode == sqlexecute.ssl.CERT_REQUIRED
    assert ctx.loaded_cert_chain == ('/tmp/client-cert.pem', '/tmp/client-key.pem')
    assert ctx.cipher_string == 'ECDHE-RSA-AES256-GCM-SHA384'
    assert ctx.minimum_version == sqlexecute.ssl.TLSVersion.TLSv1_3
    assert ctx.maximum_version == sqlexecute.ssl.TLSVersion.TLSv1_3


@pytest.mark.parametrize(
    ('tls_version', 'expected_version'),
    (
        ('TLSv1', sqlexecute.ssl.TLSVersion.TLSv1),
        ('TLSv1.1', sqlexecute.ssl.TLSVersion.TLSv1_1),
        ('TLSv1.2', sqlexecute.ssl.TLSVersion.TLSv1_2),
    ),
)
def test_create_ssl_ctx_supports_legacy_tls_version_overrides(monkeypatch, tls_version: str, expected_version) -> None:
    executor = make_executor_for_run_tests()
    ctx = FakeSSLContext()

    monkeypatch.setattr(sqlexecute.ssl, 'create_default_context', lambda **_kwargs: ctx)

    result = executor._create_ssl_ctx({'tls_version': tls_version})

    assert result is ctx
    assert ctx.minimum_version == expected_version
    assert ctx.maximum_version == expected_version


def test_create_ssl_ctx_logs_invalid_tls_version_and_keeps_default_minimum(monkeypatch, caplog) -> None:
    executor = make_executor_for_run_tests()
    ctx = FakeSSLContext()

    monkeypatch.setattr(sqlexecute.ssl, 'create_default_context', lambda **_kwargs: ctx)

    with caplog.at_level('ERROR', logger='mycli.sqlexecute'):
        result = executor._create_ssl_ctx({'tls_version': 'SSLv3'})

    assert result is ctx
    assert ctx.minimum_version == sqlexecute.ssl.TLSVersion.TLSv1_2
    assert ctx.maximum_version is None
    assert 'Invalid tls version: SSLv3' in caplog.text


def test_close_calls_connection_close_when_present() -> None:
    conn = DummyConnection(server_version='8.0.0')
    executor = make_executor_for_run_tests(conn)

    executor.close()

    assert conn.close_calls == 1


def test_close_swallows_pymysql_errors() -> None:
    conn = DummyConnection(server_version='8.0.0', close_error=pymysql.err.Error())
    executor = make_executor_for_run_tests(conn)

    executor.close()

    assert conn.close_calls == 1


def test_close_does_nothing_when_connection_is_none() -> None:
    executor = make_executor_for_run_tests()

    executor.close()



# type: ignore

from pathlib import Path

from mycli.packages.ptoolkit import history as history_module
from mycli.packages.ptoolkit.history import FileHistoryWithTimestamp


def test_file_history_with_timestamp_sets_filename(tmp_path: Path) -> None:
    history_path = tmp_path / 'history.txt'

    history = FileHistoryWithTimestamp(history_path)

    assert history.filename == history_path


def test_append_string_caches_and_stores_non_password_statement(tmp_path: Path, monkeypatch) -> None:
    history = FileHistoryWithTimestamp(tmp_path / 'history.txt')
    stored: list[str] = []
    monkeypatch.setattr(history, 'store_string', stored.append)

    history.append_string('SELECT 1')

    assert history.get_strings()[0] == 'SELECT 1'
    assert stored == ['SELECT 1']


def test_append_string_does_not_store_password_change(tmp_path: Path, monkeypatch) -> None:
    history = FileHistoryWithTimestamp(tmp_path / 'history.txt')
    stored: list[str] = []
    monkeypatch.setattr(history, 'store_string', stored.append)
    monkeypatch.setattr(history_module, 'is_password_change', lambda string: True)

    history.append_string("SET PASSWORD = 'secret'")

    assert history.get_strings()[0] == "SET PASSWORD = 'secret'"
    assert stored == []


def test_load_history_with_timestamp_returns_empty_when_file_is_missing(tmp_path: Path) -> None:
    history = FileHistoryWithTimestamp(tmp_path / 'missing-history.txt')

    assert history.load_history_with_timestamp() == []


def test_load_history_with_timestamp_parses_and_reverses_entries(tmp_path: Path) -> None:
    history_path = tmp_path / 'history.txt'
    history_path.write_text(
        '# 2026-04-02 10:00:00\n+SELECT 1\n+FROM dual\n\n# 2026-04-02 11:00:00\n+SHOW DATABASES\n',
        encoding='utf-8',
    )

    history = FileHistoryWithTimestamp(history_path)

    assert history.load_history_with_timestamp() == [
        ('SHOW DATABASES', '2026-04-02 11:00:00'),
        ('SELECT 1\nFROM dual', '2026-04-02 10:00:00'),
    ]


def test_load_history_with_timestamp_ignores_empty_separator_blocks(tmp_path: Path) -> None:
    history_path = tmp_path / 'history.txt'
    history_path.write_text(
        '# 2026-04-02 10:00:00\n\n# 2026-04-02 11:00:00\n+SELECT 1\n\ngarbage separator\n',
        encoding='utf-8',
    )

    history = FileHistoryWithTimestamp(history_path)

    assert history.load_history_with_timestamp() == [
        ('SELECT 1', '2026-04-02 11:00:00'),
    ]



# type: ignore

from unittest.mock import MagicMock

from pymysql import ProgrammingError

from mycli.packages.completion_engine import suggest_type
from mycli.packages.special import dbcommands
from mycli.packages.special.dbcommands import list_databases, list_tables, status
from test.pytests.test_completion_engine import sorted_dicts


class FakeConnection:
    def __init__(
        self,
        *,
        host: str = 'db.example',
        port: int = 3306,
        host_info: str = 'Localhost via UNIX socket',
        unix_socket: str | None = None,
        thread_id_value: int = 42,
    ) -> None:
        self.host = host
        self.port = port
        self.host_info = host_info
        self.unix_socket = unix_socket
        self._thread_id_value = thread_id_value

    def thread_id(self) -> int:
        return self._thread_id_value


class FakeCursor:
    def __init__(
        self,
        *,
        query_results: dict[str, dict[str, object]],
        connection: FakeConnection | None = None,
        fail_on_queries: set[str] | None = None,
    ) -> None:
        self.query_results = query_results
        self.connection = connection or FakeConnection()
        self.fail_on_queries = fail_on_queries or set()
        self.description = None
        self.current_query = None
        self.executed: list[str] = []

    def execute(self, query: str) -> None:
        self.executed.append(query)
        self.current_query = query
        if query in self.fail_on_queries:
            raise ProgrammingError()
        self.description = self.query_results.get(query, {}).get('description')

    def fetchall(self):
        return self.query_results.get(self.current_query, {}).get('rows', [])

    def fetchone(self):
        rows = self.query_results.get(self.current_query, {}).get('rows', [])
        return rows[0] if rows else None


def test_list_tables_verbose_preserves_field_results():
    """Test that \\dt+ table_name returns SHOW FIELDS results, not SHOW CREATE TABLE results.

    This is a regression test for a bug where the cursor was reused for SHOW CREATE TABLE,
    which overwrote the SHOW FIELDS results.
    """
    # Mock cursor that simulates MySQL behavior
    cur = MagicMock()

    # Track which query is being executed
    query_results = {
        'SHOW FIELDS FROM test_table': {
            'description': [('Field',), ('Type',), ('Null',), ('Key',), ('Default',), ('Extra',)],
            'rows': [
                ('id', 'int', 'NO', 'PRI', None, 'auto_increment'),
                ('name', 'varchar(255)', 'YES', '', None, ''),
            ],
        },
        'SHOW CREATE TABLE test_table': {
            'description': [('Table',), ('Create Table',)],
            'rows': [('test_table', 'CREATE TABLE `test_table` ...')],
        },
    }

    current_query = [None]  # Use list to allow mutation in nested function

    def execute_side_effect(query):
        current_query[0] = query
        cur.description = query_results[query]['description']
        cur.rowcount = len(query_results[query]['rows'])

    def fetchall_side_effect():
        return query_results[current_query[0]]['rows']

    def fetchone_side_effect():
        rows = query_results[current_query[0]]['rows']
        return rows[0] if rows else None

    cur.execute.side_effect = execute_side_effect
    cur.fetchall.side_effect = fetchall_side_effect
    cur.fetchone.side_effect = fetchone_side_effect

    # Call list_tables with command_verbosity=True (simulating \dt+ table_name)
    results = list_tables(cur, arg='test_table', command_verbosity=True)

    assert len(results) == 1
    result = results[0]

    # The header should be from SHOW FIELDS
    assert result.header == ['Field', 'Type', 'Null', 'Key', 'Default', 'Extra']

    # The results should contain the field data, not be empty
    # Convert to list if it's a cursor or iterable
    result_data = list(result.rows) if hasattr(result.rows, '__iter__') else result.rows
    assert len(result_data) == 2
    assert result_data[0][0] == 'id'
    assert result_data[1][0] == 'name'

    # The postamble should contain the CREATE TABLE statement
    assert 'CREATE TABLE' in result.postamble


def test_u_suggests_databases():
    suggestions = suggest_type("\\u ", "\\u ")
    assert sorted_dicts(suggestions) == sorted_dicts([{"type": "database"}])


def test_describe_table():
    suggestions = suggest_type("\\dt", "\\dt ")
    assert sorted_dicts(suggestions) == sorted_dicts([{"type": "table", "schema": []}, {"type": "view", "schema": []}, {"type": "schema"}])


def test_list_or_show_create_tables():
    suggestions = suggest_type("\\dt+", "\\dt+ ")
    assert sorted_dicts(suggestions) == sorted_dicts([{"type": "table", "schema": []}, {"type": "view", "schema": []}, {"type": "schema"}])


def test_list_tables_nonverbose_and_empty_result() -> None:
    cursor = FakeCursor(
        query_results={
            'SHOW TABLES': {
                'description': [('Tables_in_test',)],
            },
            'SHOW FIELDS FROM missing_table': {
                'description': None,
            },
        }
    )

    listed = list_tables(cursor)
    assert listed[0].header == ['Tables_in_test']
    assert listed[0].rows is cursor

    described = list_tables(cursor, arg='missing_table')
    assert described[0].header is None
    assert described[0].rows is None


def test_list_databases_with_and_without_description() -> None:
    cursor = FakeCursor(
        query_results={
            'SHOW DATABASES': {
                'description': [('Database',)],
            },
        }
    )

    listed = list_databases(cursor)
    assert listed[0].header == ['Database']
    assert listed[0].rows is cursor

    empty_cursor = FakeCursor(query_results={'SHOW DATABASES': {'description': None}})
    empty = list_databases(empty_cursor)
    assert empty[0].header is None
    assert empty[0].rows is None


def test_status_uses_global_queries_decodes_bytes_and_formats_stats(monkeypatch) -> None:
    monkeypatch.setattr(dbcommands, '__version__', '9.9.9')
    monkeypatch.setattr(dbcommands.platform, 'python_implementation', lambda: 'CPython')
    monkeypatch.setattr(dbcommands.platform, 'python_version', lambda: '3.14.0')
    monkeypatch.setattr(dbcommands.iocommands, 'is_pager_enabled', lambda: True)
    monkeypatch.setattr(dbcommands, 'get_ssl_cipher', lambda cur: 'TLS_AES_256_GCM_SHA384')
    monkeypatch.setattr(dbcommands, 'get_ssl_version', lambda cur: 'TLSv1.3')
    monkeypatch.setattr(dbcommands, 'format_uptime', lambda uptime: f'{uptime} seconds')
    monkeypatch.setenv('PAGER', 'less -SR')

    cursor = FakeCursor(
        connection=FakeConnection(host='tcp-host', port=3307, unix_socket=None),
        query_results={
            'SHOW GLOBAL STATUS;': {
                'rows': [
                    (b'Uptime', b'10'),
                    (b'Threads_connected', b'5'),
                    (b'Queries', b'20'),
                    (b'Slow_queries', b'1'),
                    (b'Opened_tables', b'2'),
                    (b'Flush_commands', b'3'),
                    (b'Open_tables', b'4'),
                ],
            },
            'SHOW GLOBAL VARIABLES;': {
                'rows': [
                    (b'version', b'8.0.0'),
                    (b'version_comment', b'Community'),
                    (b'protocol_version', b'10'),
                ],
            },
            'SELECT DATABASE(), USER();': {
                'rows': [('test_db', 'test_user')],
            },
            'SHOW SESSION VARIABLES;': {
                'rows': [
                    (b'character_set_server', b'utf8mb4'),
                    (b'character_set_database', b'utf8mb4'),
                    (b'character_set_client', b'utf8mb4'),
                    (b'character_set_connection', b'utf8mb4'),
                    (b'character_set_results', b'utf8mb4'),
                ],
            },
        },
    )

    result = status(cursor)[0]

    assert 'mycli 9.9.9 running on CPython 3.14.0' in result.preamble
    assert ('Connection id:', 42) in result.rows
    assert ('Current database:', 'test_db') in result.rows
    assert ('Current user:', 'test_user') in result.rows
    assert ('Current pager:', 'less -SR') in result.rows
    assert ('Server version:', '8.0.0 Community') in result.rows
    assert ('Protocol version:', '10') in result.rows
    assert ('SSL:', 'Cipher in use is TLS_AES_256_GCM_SHA384') in result.rows
    assert ('SSL/TLS version:', 'TLSv1.3') in result.rows
    assert ('Connection:', 'tcp-host via TCP/IP') in result.rows
    assert ('TCP port:', 3307) in result.rows
    assert ('Uptime:', '10 seconds') in result.rows
    assert 'Connections: 5' in result.postamble
    assert 'Queries: 20' in result.postamble
    assert 'Slow queries: 1' in result.postamble
    assert 'Flush tables: 3' in result.postamble
    assert 'Queries per second avg: 2.000' in result.postamble
    assert '--------------' in result.postamble


def test_status_falls_back_to_show_status_and_handles_empty_selects(monkeypatch) -> None:
    monkeypatch.setattr(dbcommands, '__version__', '1.0.0')
    monkeypatch.setattr(dbcommands.platform, 'python_implementation', lambda: 'CPython')
    monkeypatch.setattr(dbcommands.platform, 'python_version', lambda: '3.10.0')
    monkeypatch.setattr(dbcommands.iocommands, 'is_pager_enabled', lambda: False)
    monkeypatch.setattr(dbcommands, 'get_ssl_version', lambda cur: 'none')
    monkeypatch.setattr(dbcommands, 'format_uptime', lambda uptime: f'{uptime} seconds')

    cursor = FakeCursor(
        connection=FakeConnection(unix_socket='/tmp/mysql.sock'),
        fail_on_queries={'SHOW GLOBAL STATUS;'},
        query_results={
            'SHOW STATUS;': {
                'rows': [
                    ('Slow_queries', '0'),
                    ('Opened_tables', '1'),
                    ('Open_tables', '2'),
                ],
            },
            'SHOW GLOBAL VARIABLES;': {
                'rows': [
                    ('version', '5.7.0'),
                    ('version_comment', 'Server'),
                    ('protocol_version', '10'),
                    ('socket', '/tmp/mysql.sock'),
                ],
            },
            'SHOW SESSION VARIABLES;': {
                'rows': [],
            },
            'SELECT DATABASE(), USER();': {
                'rows': [],
            },
        },
    )

    result = status(cursor)[0]

    assert cursor.executed[0:2] == ['SHOW GLOBAL STATUS;', 'SHOW STATUS;']
    assert ('Current database:', '') in result.rows
    assert ('Current user:', '') in result.rows
    assert ('Current pager:', 'stdout') in result.rows
    assert ('Connection:', 'Localhost via UNIX socket') in result.rows
    assert ('UNIX socket:', '/tmp/mysql.sock') in result.rows
    assert ('Server characterset:', '') in result.rows
    assert ('Db     characterset:', '') in result.rows
    assert ('Client characterset:', '') in result.rows
    assert ('Conn.  characterset:', '') in result.rows
    assert 'Connections:' not in result.postamble
    assert '--------------' in result.postamble


def test_status_uses_system_default_pager_when_enabled_without_env(monkeypatch) -> None:
    monkeypatch.setattr(dbcommands.iocommands, 'is_pager_enabled', lambda: True)
    monkeypatch.setattr(dbcommands, 'get_ssl_version', lambda cur: 'TLS')
    monkeypatch.setattr(dbcommands.platform, 'python_implementation', lambda: 'CPython')
    monkeypatch.setattr(dbcommands.platform, 'python_version', lambda: '3.14.0')
    monkeypatch.delenv('PAGER', raising=False)

    cursor = FakeCursor(
        query_results={
            'SHOW GLOBAL STATUS;': {
                'rows': [('Slow_queries', '0'), ('Opened_tables', '1'), ('Open_tables', '2')],
            },
            'SHOW GLOBAL VARIABLES;': {
                'rows': [('version', '8.0'), ('version_comment', 'Server'), ('protocol_version', '10')],
            },
            'SELECT DATABASE(), USER();': {
                'rows': [('db', 'user')],
            },
            'SHOW SESSION VARIABLES;': {
                'rows': [
                    ('character_set_server', 'utf8'),
                    ('character_set_database', 'utf8'),
                    ('character_set_client', 'utf8'),
                    ('character_set_connection', 'utf8'),
                    ('character_set_results', 'utf8'),
                ],
            },
        },
    )

    result = status(cursor)[0]

    assert ('Current pager:', 'System default') in result.rows



from prompt_toolkit.application import get_app
from prompt_toolkit.enums import DEFAULT_BUFFER
from prompt_toolkit.filters import Condition, Filter

from mycli.packages.special import iocommands
from mycli.packages.special.main import COMMANDS as SPECIAL_COMMANDS


def cli_is_multiline(mycli) -> Filter:
    @Condition
    def cond():
        if not mycli.multi_line:
            return False
        else:
            doc = get_app().layout.get_buffer_by_name(DEFAULT_BUFFER).document
            return not _multiline_exception(doc.text)

    return cond


def _multiline_exception(text: str) -> bool:
    orig = text
    text = text.strip()
    first_word = text.split()[0] if text else ''

    # Multi-statement favorite query is a special case. Because there will
    # be a semicolon separating statements, we can't consider semicolon an
    # EOL. Let's consider an empty line an EOL instead.
    if first_word.startswith("\\fs"):
        return orig.endswith("\n")

    return (
        # Special Command
        first_word.startswith("\\")
        or text.endswith((
            # Ended with the current delimiter (usually a semi-column)
            iocommands.get_current_delimiter(),
            # or ended with certain commands
            "\\g",
            "\\G",
            r"\e",
            r"\edit",
            r"\clip",
        ))
        or
        # non-backslashed special commands such as "exit" or "help" don't need semicolon
        first_word in SPECIAL_COMMANDS
        or
        # uppercase variants accepted
        first_word.lower() in SPECIAL_COMMANDS
        or
        # just a plain enter without any text
        (first_word == "")
    )



import threading
from typing import Callable

import pymysql

from mycli.packages.special.main import COMMANDS
from mycli.packages.sqlresult import SQLResult
from mycli.sqlcompleter import SQLCompleter
from mycli.sqlexecute import ServerSpecies, SQLExecute


class CompletionRefresher:
    refreshers: dict = {}

    def __init__(self) -> None:
        self._completer_thread: threading.Thread | None = None
        self._restart_refresh = threading.Event()

    def refresh(
        self,
        executor: SQLExecute,
        callbacks: Callable | list[Callable],
        completer_options: dict | None = None,
    ) -> list[SQLResult]:
        """Creates a SQLCompleter object and populates it with the relevant
        completion suggestions in a background thread.

        executor - SQLExecute object, used to extract the credentials to connect
                   to the database.
        callbacks - A function or a list of functions to call after the thread
                    has completed the refresh. The newly created completion
                    object will be passed in as an argument to each callback.
        completer_options - dict of options to pass to SQLCompleter.

        """
        if completer_options is None:
            completer_options = {}

        if self.is_refreshing():
            self._restart_refresh.set()
            return [SQLResult(status="Auto-completion refresh restarted.")]
        else:
            self._completer_thread = threading.Thread(
                target=self._bg_refresh, args=(executor, callbacks, completer_options), name="completion_refresh"
            )
            self._completer_thread.daemon = True
            self._completer_thread.start()
            return [SQLResult(status="Auto-completion refresh started in the background.")]

    def is_refreshing(self) -> bool:
        return bool(self._completer_thread and self._completer_thread.is_alive())

    def _bg_refresh(
        self,
        sqlexecute: SQLExecute,
        callbacks: Callable | list[Callable],
        completer_options: dict,
    ) -> None:
        completer = SQLCompleter(**completer_options)

        # Create a new sqlexecute method to populate the completions.
        e = sqlexecute
        try:
            executor = SQLExecute(
                e.dbname,
                e.user,
                e.password,
                e.host,
                e.port,
                e.socket,
                e.character_set,
                e.local_infile,
                e.ssl,
                e.ssh_user,
                e.ssh_host,
                e.ssh_port,
                e.ssh_password,
                e.ssh_key_filename,
            )
        except pymysql.err.OperationalError:
            return

        # If callbacks is a single function then push it into a list.
        if callable(callbacks):
            callbacks = [callbacks]

        while 1:
            for refresher in self.refreshers.values():
                refresher(completer, executor)
                if self._restart_refresh.is_set():
                    self._restart_refresh.clear()
                    break
            else:
                # Break out of while loop if the for loop finishes natually
                # without hitting the break statement.
                break

            # Start over the refresh from the beginning if the for loop hit the
            # break statement.
            continue

        for callback in callbacks:
            callback(completer)

        executor.close()


def refresher(name: str, refreshers: dict = CompletionRefresher.refreshers) -> Callable:
    """Decorator to add the decorated function to the dictionary of
    refreshers. Any function decorated with a @refresher will be executed as
    part of the completion refresh routine."""

    def wrapper(wrapped):
        refreshers[name] = wrapped
        return wrapped

    return wrapper


@refresher("databases")
def refresh_databases(completer: SQLCompleter, executor: SQLExecute) -> None:
    completer.extend_database_names(executor.databases())


@refresher("schemata")
def refresh_schemata(completer: SQLCompleter, executor: SQLExecute) -> None:
    # schemata - In MySQL Schema is the same as database. But for mycli
    # schemata will be the name of the current database.
    completer.extend_schemata(executor.dbname)
    completer.set_dbname(executor.dbname)


@refresher("tables")
def refresh_tables(completer: SQLCompleter, executor: SQLExecute) -> None:
    table_columns_dbresult = list(executor.table_columns())
    completer.extend_relations(table_columns_dbresult, kind="tables")
    completer.extend_columns(table_columns_dbresult, kind="tables")


@refresher("foreign_keys")
def refresh_foreign_keys(completer: SQLCompleter, executor: SQLExecute) -> None:
    completer.extend_foreign_keys(executor.foreign_keys())


@refresher("enum_values")
def refresh_enum_values(completer: SQLCompleter, executor: SQLExecute) -> None:
    completer.extend_enum_values(executor.enum_values())


@refresher("users")
def refresh_users(completer: SQLCompleter, executor: SQLExecute) -> None:
    completer.extend_users(executor.users())


# @refresher('views')
# def refresh_views(completer: SQLCompleter, executor: SQLExecute) -> None:
#     completer.extend_relations(executor.views(), kind='views')
#     completer.extend_columns(executor.view_columns(), kind='views')


@refresher("functions")
def refresh_functions(completer: SQLCompleter, executor: SQLExecute) -> None:
    completer.extend_functions(executor.functions())
    if executor.server_info and executor.server_info.species == ServerSpecies.TiDB:
        completer.extend_functions(completer.tidb_functions, builtin=True)


@refresher("procedures")
def refresh_procedures(completer: SQLCompleter, executor: SQLExecute) -> None:
    completer.extend_procedures(executor.procedures())


@refresher("character_sets")
def refresh_character_sets(completer: SQLCompleter, executor: SQLExecute) -> None:
    completer.extend_character_sets(executor.character_sets())


@refresher("collations")
def refresh_collations(completer: SQLCompleter, executor: SQLExecute) -> None:
    completer.extend_collations(executor.collations())


@refresher("special_commands")
def refresh_special(completer: SQLCompleter, executor: SQLExecute) -> None:
    completer.extend_special_commands(list(COMMANDS.keys()))


@refresher("show_commands")
def refresh_show_commands(completer: SQLCompleter, executor: SQLExecute) -> None:
    completer.extend_show_items(executor.show_candidates())


@refresher("keywords")
def refresh_keywords(completer: SQLCompleter, executor: SQLExecute) -> None:
    if executor.server_info and executor.server_info.species == ServerSpecies.TiDB:
        completer.extend_keywords(completer.tidb_keywords, replace=True)



# type: ignore

import os
from textwrap import dedent

from behave import then, when
import wrappers


@when("we start external editor providing a file name")
def step_edit_file(context):
    """Edit file with external editor."""
    context.editor_file_name = os.path.join(context.package_root, f"test_file_{context.conf['vi']}.sql")
    if os.path.exists(context.editor_file_name):
        os.remove(context.editor_file_name)
    context.cli.sendline(f"\\e {os.path.basename(context.editor_file_name)}")
    wrappers.expect_exact(context, 'Entering Ex mode.  Type "visual" to go to Normal mode.', timeout=4)
    wrappers.expect_exact(context, "\r\n:", timeout=4)


@when('we type "{query}" in the editor')
def step_edit_type_sql(context, query):
    context.cli.sendline("i")
    context.cli.sendline(query)
    context.cli.sendline(".")
    wrappers.expect_exact(context, "\r\n:", timeout=4)


@when("we exit the editor")
def step_edit_quit(context):
    context.cli.sendline("x")
    wrappers.expect_exact(context, "written", timeout=4)


@then('we see "{query}" in prompt')
def step_edit_done_sql(context, query):
    for match in query.split(" "):
        wrappers.expect_exact(context, match, timeout=5)
    # Cleanup the command line.
    context.cli.sendcontrol("c")
    # Cleanup the edited file.
    if context.editor_file_name and os.path.exists(context.editor_file_name):
        os.remove(context.editor_file_name)


@when("we tee output")
def step_tee_ouptut(context):
    context.tee_file_name = os.path.join(context.package_root, f"tee_file_{context.conf['vi']}.sql")
    if os.path.exists(context.tee_file_name):
        os.remove(context.tee_file_name)
    context.cli.sendline(f"tee {os.path.basename(context.tee_file_name)}")


@when('we select "select {param}"')
def step_query_select_number(context, param):
    context.cli.sendline(f"select {param}")
    expected = (
        dedent(
            f"""
            +{'-' * (len(param) + 2)}+\r
            | {param} |\r
            +{'-' * (len(param) + 2)}+\r
            | {param} |\r
            +{'-' * (len(param) + 2)}+
            """
        ).strip()
        + '\r\n\r\n'
    )

    wrappers.expect_pager(
        context,
        expected,
        timeout=5,
    )
    wrappers.expect_exact(context, "1 row in set", timeout=2)


@then('we see tabular result "{result}"')
def step_see_tabular_result(context, result):
    wrappers.expect_exact(context, f'| {result} |', timeout=2)


@then('we see csv result "{result}"')
def step_see_csv_result(context, result):
    wrappers.expect_exact(context, f'"{result}"', timeout=2)


@when('we query "{query}"')
def step_query(context, query):
    context.cli.sendline(query)


@when("we notee output")
def step_notee_output(context):
    context.cli.sendline("notee")


@then("we see 123456 in tee output")
def step_see_123456_in_ouput(context):
    with open(context.tee_file_name) as f:
        assert "123456" in f.read()
    if os.path.exists(context.tee_file_name):
        os.remove(context.tee_file_name)


@then('we see csv {result} in file output')
def step_see_csv_result_in_redirected_ouput(context, result):
    wrappers.expect_exact(context, f'"{result}"', timeout=2)
    temp_filename = "/tmp/output1.csv"
    if os.path.exists(temp_filename):
        os.remove(temp_filename)


@then('we see text {result} in file output')
def step_see_text_result_in_redirected_ouput(context, result):
    wrappers.expect_exact(context, f' {result}', timeout=2)
    temp_filename = "/tmp/output1.txt"
    if os.path.exists(temp_filename):
        os.remove(temp_filename)


@then("we see space 12 in command output")
def step_see_space_12_in_command_ouput(context):
    wrappers.expect_exact(context, ' 12', timeout=2)


@then("we see space 6 in command output")
def step_see_space_6_in_command_ouput(context):
    wrappers.expect_exact(context, ' 6', timeout=2)


@then('delimiter is set to "{delimiter}"')
def delimiter_is_set(context, delimiter):
    wrappers.expect_exact(context, f"Changed delimiter to {delimiter}", timeout=2)



import builtins
import importlib
from types import SimpleNamespace
from typing import Any, cast
from unittest.mock import patch

import click
import pytest

from mycli.packages.special import llm as llm_module
from mycli.packages.special.llm import (
    NEED_DEPENDENCIES,
    USAGE,
    _build_command_tree,
    build_command_tree,
    ensure_mycli_template,
    get_completions,
    get_sample_data,
    get_schema,
    handle_llm,
    is_llm_command,
    run_external_cmd,
    sql_using_llm,
    truncate_list_elements,
    truncate_table_lines,
)
from mycli.packages.special.main import COMMANDS
from mycli.packages.sqlresult import SQLResult


# Override executor fixture to avoid real DB connections during llm tests
@pytest.fixture
def executor():
    """Dummy executor fixture"""
    return None


def test_reload_llm_module_handles_disabled_and_import_error_paths(monkeypatch) -> None:
    with monkeypatch.context() as m:
        m.setenv("MYCLI_LLM_OFF", "1")
        importlib.reload(llm_module)
        assert llm_module.LLM_IMPORTED is False
        assert llm_module.LLM_CLI_IMPORTED is False

    importlib.reload(llm_module)

    original_import = builtins.__import__

    def fake_import(name, globals=None, locals=None, fromlist=(), level=0):  # noqa: A002
        if name == "llm" or name == "llm.cli":
            raise ImportError("no llm")
        return original_import(name, globals, locals, fromlist, level)

    with monkeypatch.context() as m:
        m.delenv("MYCLI_LLM_OFF", raising=False)
        m.setattr(builtins, "__import__", fake_import)
        importlib.reload(llm_module)
        assert llm_module.LLM_IMPORTED is False
        assert llm_module.LLM_CLI_IMPORTED is False

    importlib.reload(llm_module)


def test_reload_llm_module_handles_cli_import_error(monkeypatch) -> None:
    original_import = builtins.__import__

    def fake_import(name, globals=None, locals=None, fromlist=(), level=0):  # noqa: A002
        if name == "llm.cli":
            raise ImportError("no llm cli")
        return original_import(name, globals, locals, fromlist, level)

    with monkeypatch.context() as m:
        m.delenv("MYCLI_LLM_OFF", raising=False)
        m.setattr(builtins, "__import__", fake_import)
        importlib.reload(llm_module)
        assert llm_module.LLM_IMPORTED is True
        assert llm_module.LLM_CLI_IMPORTED is False

    importlib.reload(llm_module)


def test_build_command_tree_handles_groups_models_and_leaf(monkeypatch) -> None:
    monkeypatch.setattr(
        llm_module,
        "llm",
        SimpleNamespace(get_models=lambda: [SimpleNamespace(model_id="gpt-4o"), SimpleNamespace(model_id="llama3")]),
        raising=False,
    )

    models_group = click.Group("models")
    models_group.add_command(click.Command("default"))
    root = click.Group("root")
    root.add_command(click.Command("prompt"))
    root.add_command(models_group)

    assert _build_command_tree(root) == {
        "prompt": None,
        "models": {"default": {"gpt-4o": None, "llama3": None}},
    }
    assert build_command_tree(click.Command("leaf")) == {}


def test_get_completions_walks_tree_and_skips_flags() -> None:
    tree = {
        "models": {"default": {"gpt-4o": None}},
        "prompt": None,
    }

    assert get_completions([], tree) == ["models", "prompt"]
    assert get_completions(["models"], tree) == ["default"]
    assert get_completions(["models", "--help"], tree) == ["default"]
    assert get_completions(["models", "default"], tree) == ["gpt-4o"]
    assert get_completions(["missing"], tree) == []
    assert get_completions(["prompt"], tree) == []


def test_cli_commands_is_cached(monkeypatch) -> None:
    llm_module.cli_commands.cache_clear()
    monkeypatch.setattr(llm_module, "cli", SimpleNamespace(commands={"models": object(), "prompt": object()}))

    assert llm_module.cli_commands() == ["models", "prompt"]

    monkeypatch.setattr(llm_module, "cli", SimpleNamespace(commands={"install": object()}))
    assert llm_module.cli_commands() == ["models", "prompt"]
    llm_module.cli_commands.cache_clear()


def test_run_external_cmd_capture_output_and_restore_argv(monkeypatch, capsys) -> None:
    original_argv = list(llm_module.sys.argv)

    def fake_run_module(cmd: str, run_name: str) -> None:
        assert cmd == "llm"
        assert run_name == "__main__"
        print("stdout text")
        llm_module.sys.stderr.write("stderr text")

    monkeypatch.setattr(llm_module, "run_module", fake_run_module)

    code, output = run_external_cmd("llm", "models", capture_output=True)

    assert code == 0
    assert "stdout text" in output
    assert "stderr text" in output
    assert llm_module.sys.argv == original_argv
    captured = capsys.readouterr()
    assert captured.out == ""
    assert captured.err == ""


def test_run_external_cmd_nonzero_exit_raises_with_output(monkeypatch) -> None:
    def fake_run_module(cmd: str, run_name: str) -> None:
        print("failed output")
        raise SystemExit(2)

    monkeypatch.setattr(llm_module, "run_module", fake_run_module)

    with pytest.raises(RuntimeError, match="failed output"):
        run_external_cmd("llm", capture_output=True)


def test_run_external_cmd_nonzero_exit_raises_without_output(monkeypatch) -> None:
    monkeypatch.setattr(llm_module, "run_module", lambda cmd, run_name: (_ for _ in ()).throw(SystemExit(3)))

    with pytest.raises(RuntimeError, match=r"Command llm failed with exit code 3\."):
        run_external_cmd("llm")


def test_run_external_cmd_exception_paths_and_restart(monkeypatch) -> None:
    monkeypatch.setattr(llm_module, "run_module", lambda cmd, run_name: (_ for _ in ()).throw(ValueError("boom")))

    with pytest.raises(RuntimeError, match=r"Command llm failed: boom"):
        run_external_cmd("llm")

    def fake_run_module_capture(cmd: str, run_name: str) -> None:
        print("capture boom")
        raise ValueError("boom")

    monkeypatch.setattr(llm_module, "run_module", fake_run_module_capture)
    with pytest.raises(RuntimeError, match="capture boom"):
        run_external_cmd("llm", capture_output=True)

    execv_calls: list[tuple[str, list[str]]] = []
    monkeypatch.setattr(llm_module, "run_module", lambda cmd, run_name: (_ for _ in ()).throw(SystemExit(0)))
    monkeypatch.setattr(llm_module.os, "execv", lambda exe, args: execv_calls.append((exe, args)))

    code, output = run_external_cmd("llm", "install", restart_cli=True)

    assert code == 0
    assert output == ""
    assert execv_calls == [(llm_module.sys.executable, [llm_module.sys.executable] + llm_module.sys.argv)]


def test_ensure_mycli_template_returns_early_or_replaces(monkeypatch) -> None:
    calls: list[tuple] = []

    def fake_run_external_cmd(*args, **kwargs):
        calls.append((args, kwargs))
        return (0, "")

    monkeypatch.setattr(llm_module, "run_external_cmd", fake_run_external_cmd)
    ensure_mycli_template()

    assert calls == [
        (("llm", "templates", "show", llm_module.LLM_TEMPLATE_NAME), {"capture_output": True, "raise_exception": False}),
    ]

    calls.clear()

    def fake_run_external_cmd_missing(*args, **kwargs):
        calls.append((args, kwargs))
        return (1, "") if len(calls) == 1 else (0, "")

    monkeypatch.setattr(llm_module, "run_external_cmd", fake_run_external_cmd_missing)
    ensure_mycli_template()

    assert calls == [
        (("llm", "templates", "show", llm_module.LLM_TEMPLATE_NAME), {"capture_output": True, "raise_exception": False}),
        (("llm", llm_module.PROMPT, "--save", llm_module.LLM_TEMPLATE_NAME), {}),
    ]

    calls.clear()
    monkeypatch.setattr(llm_module, "run_external_cmd", fake_run_external_cmd)
    ensure_mycli_template(replace=True)

    assert calls == [
        (("llm", llm_module.PROMPT, "--save", llm_module.LLM_TEMPLATE_NAME), {}),
    ]


@patch("mycli.packages.special.llm.llm")
def test_llm_command_without_args(mock_llm, executor):
    r"""
    Invoking \llm without any arguments should print the usage and raise FinishIteration.
    """
    assert mock_llm is not None
    test_text = r"\llm"
    with pytest.raises(llm_module.FinishIteration) as exc_info:
        handle_llm(test_text, executor, 'mysql', 0, 0)
    # Should return usage message when no args provided
    assert exc_info.value.results == [SQLResult(preamble=USAGE)]


@patch("mycli.packages.special.llm.llm")
def test_llm_command_with_help_subcommand(mock_llm, executor):
    r"""
    Invoking \llm with "help" should print the usage and raise FinishIteration.
    """
    assert mock_llm is not None
    test_text = r"\llm help"
    with pytest.raises(llm_module.FinishIteration) as exc_info:
        handle_llm(test_text, executor, 'mysql', 0, 0)
    # Should return usage message when "help" subcommand or variant is provided
    assert exc_info.value.results == [SQLResult(preamble=USAGE)]


@patch("mycli.packages.special.llm.llm")
@patch("mycli.packages.special.llm.run_external_cmd")
def test_llm_command_with_c_flag(mock_run_cmd, mock_llm, executor):
    string = "Hello, no SQL today."
    # Suppose the LLM returns some text without fenced SQL
    mock_run_cmd.return_value = (0, string)
    test_text = r"\llm -c 'Something?'"
    with pytest.raises(llm_module.FinishIteration) as exc_info:
        handle_llm(test_text, executor, 'mysql', 0, 0)
    # Expect raw output when no SQL fence found
    assert exc_info.value.results == [SQLResult(preamble=string)]


@patch("mycli.packages.special.llm.llm")
@patch("mycli.packages.special.llm.run_external_cmd")
def test_llm_command_with_c_flag_and_fenced_sql(mock_run_cmd, mock_llm, executor):
    # Return text containing a fenced SQL block
    sql_text = "SELECT * FROM users;"
    fenced = f"Here you go:\n```sql\n{sql_text}\n```"
    mock_run_cmd.return_value = (0, fenced)
    test_text = r"\llm -c 'Rewrite SQL'"
    result, sql, duration = handle_llm(test_text, executor, 'mysql', 0, 0)
    # Without verbosity, result is empty, sql extracted
    assert sql == sql_text
    assert result == ""
    assert isinstance(duration, float)


@patch("mycli.packages.special.llm.llm")
@patch("mycli.packages.special.llm.run_external_cmd")
def test_llm_command_known_subcommand(mock_run_cmd, mock_llm, executor):
    # 'models' is a known subcommand
    test_text = r"\llm models"
    with pytest.raises(llm_module.FinishIteration) as exc_info:
        handle_llm(test_text, executor, 'mysql', 0, 0)
    mock_run_cmd.assert_called_once_with("llm", "models", restart_cli=False)
    assert exc_info.value.results is None


@patch("mycli.packages.special.llm.llm")
@patch("mycli.packages.special.llm.run_external_cmd")
def test_llm_command_with_help_flag(mock_run_cmd, mock_llm, executor):
    test_text = r"\llm --help"
    with pytest.raises(llm_module.FinishIteration) as exc_info:
        handle_llm(test_text, executor, 'mysql', 0, 0)
    mock_run_cmd.assert_called_once_with("llm", "--help", restart_cli=False)
    assert exc_info.value.results is None


@patch("mycli.packages.special.llm.llm")
@patch("mycli.packages.special.llm.run_external_cmd")
def test_llm_command_with_install_flag(mock_run_cmd, mock_llm, executor):
    test_text = r"\llm install openai"
    with pytest.raises(llm_module.FinishIteration) as exc_info:
        handle_llm(test_text, executor, 'mysql', 0, 0)
    mock_run_cmd.assert_called_once_with("llm", "install", "openai", restart_cli=True)
    assert exc_info.value.results is None


@patch("mycli.packages.special.llm.llm")
@patch("mycli.packages.special.llm.ensure_mycli_template")
@patch("mycli.packages.special.llm.sql_using_llm")
def test_llm_command_with_prompt(mock_sql_using_llm, mock_ensure_template, mock_llm, executor):
    r"""
    \llm prompt 'question' should use template and call sql_using_llm
    """
    mock_sql_using_llm.return_value = ("CTX", "SELECT 1;")
    test_text = r"\llm prompt 'Test?'"
    context, sql, duration = handle_llm(test_text, executor, 'mysql', 0, 0)
    mock_ensure_template.assert_called_once()
    mock_sql_using_llm.assert_called()
    assert context == "CTX"
    assert sql == "SELECT 1;"
    assert isinstance(duration, float)


@patch("mycli.packages.special.llm.llm")
@patch("mycli.packages.special.llm.ensure_mycli_template")
@patch("mycli.packages.special.llm.sql_using_llm")
def test_llm_command_question_with_context(mock_sql_using_llm, mock_ensure_template, mock_llm, executor):
    r"""
    \llm 'question' treats as prompt and returns SQL
    """
    mock_sql_using_llm.return_value = ("CTX2", "SELECT 2;")
    test_text = r"\llm 'Top 10?'"
    context, sql, duration = handle_llm(test_text, executor, 'mysql', 0, 0)
    mock_ensure_template.assert_called_once()
    mock_sql_using_llm.assert_called()
    assert context == "CTX2"
    assert sql == "SELECT 2;"
    assert isinstance(duration, float)


@patch("mycli.packages.special.llm.llm")
@patch("mycli.packages.special.llm.ensure_mycli_template")
@patch("mycli.packages.special.llm.sql_using_llm")
def test_llm_command_question_verbose(mock_sql_using_llm, mock_ensure_template, mock_llm, executor):
    r"""
    \llm+ returns verbose context and SQL
    """
    mock_sql_using_llm.return_value = ("NO_CTX", "SELECT 42;")
    test_text = r"\llm- 'Succinct?'"
    context, sql, duration = handle_llm(test_text, executor, 'mysql', 0, 0)
    assert context == ""
    assert sql == "SELECT 42;"
    assert isinstance(duration, float)


def test_handle_llm_without_dependencies(executor, monkeypatch) -> None:
    monkeypatch.setattr(llm_module, "LLM_IMPORTED", False)

    with pytest.raises(llm_module.FinishIteration) as exc_info:
        handle_llm(r"\llm anything", executor, "mysql", 0, 0)

    assert exc_info.value.results == [SQLResult(preamble=NEED_DEPENDENCIES)]


@patch("mycli.packages.special.llm.llm")
def test_handle_llm_wraps_context_errors(mock_llm, executor, monkeypatch) -> None:
    assert mock_llm is not None
    monkeypatch.setattr(llm_module, "ensure_mycli_template", lambda: (_ for _ in ()).throw(ValueError("bad template")))

    with pytest.raises(RuntimeError, match="bad template"):
        handle_llm(r"\llm 'Top 10?'", executor, "mysql", 0, 0)


def test_is_llm_command():
    # Valid llm command variants
    for cmd in ["\\llm", "\\ai"]:
        assert is_llm_command(cmd + " 'x'")
    # Invalid commands
    assert not is_llm_command("select * from table;")


def test_sql_using_llm_no_connection():
    # Should error if no database cursor provided
    with pytest.raises(RuntimeError) as exc_info:
        sql_using_llm(None, question="test")
    assert "Connect to a database" in str(exc_info.value)


def test_truncate_list_elements_and_table_lines(monkeypatch) -> None:
    monkeypatch.setattr(llm_module.sys, "getsizeof", lambda value: len(value) if isinstance(value, (str, bytes)) else 8)

    row = ["a" * 250, b"b" * 250, 1]
    truncated = truncate_list_elements(row, prompt_field_truncate=250, prompt_section_truncate=300)
    assert truncated == ["a" * 50, b"b" * 50, 1]
    assert truncate_list_elements(row, prompt_field_truncate=0, prompt_section_truncate=0) is row
    assert truncate_list_elements(["abcdef"], prompt_field_truncate=3, prompt_section_truncate=0) == ["abc"]

    table = ["a" * 100, "b" * 100, "c" * 100]
    assert truncate_table_lines(table.copy(), prompt_section_truncate=0) == table
    assert truncate_table_lines(table.copy(), prompt_section_truncate=210) == ["a" * 100, "b" * 100]
    assert truncate_table_lines(table.copy(), prompt_section_truncate=150) == ["a" * 100]
    assert truncate_table_lines(["a" * 100], prompt_section_truncate=50) == []


def test_get_schema_and_sample_data_use_cache_and_skip_bad_rows(monkeypatch) -> None:
    llm_module.SCHEMA_DATA_CACHE.clear()
    llm_module.SAMPLE_DATA_CACHE.clear()
    monkeypatch.setattr(llm_module.click, "echo", lambda message: None)
    monkeypatch.setattr(llm_module.sys, "getsizeof", lambda value: len(value) if isinstance(value, (str, bytes)) else 8)

    class DummyCursor:
        def __init__(self) -> None:
            self.executed: list[str] = []
            self.description: list[tuple[str, None]] = []
            self._rows: list[tuple[str]] = []
            self._row: tuple[int, str] | None = None

        def execute(self, query: str) -> None:
            self.executed.append(query)
            if "information_schema.columns" in query:
                self._rows = [("orders(id int)",), ("users(name text)",)]
                return
            if query == "SHOW TABLES":
                self._rows = [("orders",), ("broken",), ("empty",)]
                return
            if "`orders`" in query:
                self.description = [("id", None), ("name", None)]
                self._row = (1, "alice")
                return
            if "`broken`" in query:
                raise RuntimeError("bad table")
            if "`empty`" in query:
                self.description = [("id", None)]
                self._row = None
                return
            raise AssertionError(f"unexpected query: {query}")

        def fetchall(self) -> list[tuple[str]]:
            return self._rows

        def fetchone(self) -> tuple[int, str] | None:
            return self._row

    cursor = DummyCursor()

    assert get_schema(cast(Any, cursor), "mysql", 0) == "orders(id int)\nusers(name text)"
    assert get_schema(cast(Any, cursor), "mysql", 0) == "orders(id int)\nusers(name text)"
    sample_data = get_sample_data(cast(Any, cursor), "mysql", 10, 100)
    assert sample_data == {"orders": [("id", 1), ("name", "alice")]}
    assert get_sample_data(cast(Any, cursor), "mysql", 10, 100) == sample_data
    assert cursor.executed.count("SHOW TABLES") == 1
    assert sum(1 for query in cursor.executed if "information_schema.columns" in query) == 1


# Test sql_using_llm with dummy cursor and fenced SQL output
@patch("mycli.packages.special.llm.run_external_cmd")
def test_sql_using_llm_success(mock_run_cmd):
    llm_module.SCHEMA_DATA_CACHE.clear()
    llm_module.SAMPLE_DATA_CACHE.clear()

    # Dummy cursor simulating database schema and sample data
    class DummyCursor:
        def __init__(self):
            self._last = []
            self.executed = []

        def execute(self, query):
            self.executed.append(query)
            if "information_schema.columns" in query:
                self._last = [("table1(col1 int,col2 text)",), ("table2(colA varchar(20))",)]
            elif query.strip().upper().startswith("SHOW TABLES"):
                self._last = [("table1",), ("table2",)]
            elif query.strip().upper().startswith("SELECT * FROM"):
                self.description = [("col1", None), ("col2", None)]
                self._row = (1, "abc")

        def fetchall(self):
            return getattr(self, "_last", [])

        def fetchone(self):
            return getattr(self, "_row", None)

    dummy_cur = DummyCursor()
    # Simulate llm CLI returning a fenced SQL result
    sql_text = "SELECT 1, 'abc';"
    fenced = f"Note\n```sql\n{sql_text}\n```"
    mock_run_cmd.return_value = (0, fenced)
    result, sql = sql_using_llm(dummy_cur, question="dummy", dbname='mysql')

    assert any("information_schema.columns" in query for query in dummy_cur.executed)
    assert "SHOW TABLES" in dummy_cur.executed
    assert any(query.strip().upper().startswith("SELECT * FROM") for query in dummy_cur.executed)
    mock_run_cmd.assert_called_once_with(
        "llm",
        "--template",
        llm_module.LLM_TEMPLATE_NAME,
        "--param",
        "db_schema",
        "table1(col1 int,col2 text)\ntable2(colA varchar(20))",
        "--param",
        "sample_data",
        {"table1": [("col1", 1), ("col2", "abc")], "table2": [("col1", 1), ("col2", "abc")]},
        "--param",
        "question",
        "dummy",
        " ",
        capture_output=True,
    )
    assert result == fenced
    assert sql == sql_text


def test_sql_using_llm_requires_schema_and_allows_missing_sql(monkeypatch) -> None:
    class DummyCursor:
        pass

    with pytest.raises(RuntimeError, match="Choose a schema and try again."):
        sql_using_llm(cast(Any, DummyCursor()), question="test", dbname="")

    monkeypatch.setattr(llm_module, "get_schema", lambda cur, dbname, truncate: "schema")
    monkeypatch.setattr(llm_module, "get_sample_data", lambda cur, dbname, field_truncate, section_truncate: {"t": [("c", 1)]})
    monkeypatch.setattr(llm_module.click, "echo", lambda message: None)
    monkeypatch.setattr(llm_module, "run_external_cmd", lambda *args, **kwargs: (0, "No fenced SQL here."))

    result, sql = sql_using_llm(cast(Any, DummyCursor()), question="test", dbname="mysql")

    assert result == "No fenced SQL here."
    assert sql == ""


# Test handle_llm supports registered command names without args
@pytest.mark.parametrize("prefix", [r"\llm", r"\ai"])
def test_handle_llm_registered_aliases_without_args(prefix, executor, monkeypatch):
    assert prefix in COMMANDS
    assert COMMANDS[prefix].handler is COMMANDS[r"\llm"].handler
    assert COMMANDS[prefix].command == r"\llm"
    monkeypatch.setattr(llm_module, "llm", object())
    with pytest.raises(llm_module.FinishIteration) as exc_info:
        handle_llm(prefix, executor, 'mysql', 0, 0)
    assert exc_info.value.results == [SQLResult(preamble=USAGE)]



import re
import shlex
from shutil import which

from prompt_toolkit import search
from prompt_toolkit.key_binding.key_processor import KeyPressEvent
from pyfzf import FzfPrompt

from mycli.packages.ptoolkit.history import FileHistoryWithTimestamp
from mycli.packages.ptoolkit.utils import safe_invalidate_display


class Fzf(FzfPrompt):
    def __init__(self):
        self.executable = which("fzf")
        if self.executable:
            super().__init__()

    def is_available(self) -> bool:
        return self.executable is not None


def search_history(
    event: KeyPressEvent,
    highlight_preview: bool = False,
    highlight_style: str = 'default',
    incremental: bool = False,
) -> None:
    buffer = event.current_buffer
    history = buffer.history

    fzf = Fzf()

    if incremental or not fzf.is_available() or not isinstance(history, FileHistoryWithTimestamp):
        # Fallback to default reverse incremental search
        search.start_search(direction=search.SearchDirection.BACKWARD)
        return

    history_items_with_timestamp = history.load_history_with_timestamp()

    formatted_history_items = []
    original_history_items = []
    seen = {}
    for item, timestamp in history_items_with_timestamp:
        formatted_item = re.sub(r'\s+', ' ', item)
        timestamp = timestamp.split(".")[0] if "." in timestamp else timestamp
        if formatted_item in seen:
            continue
        seen[formatted_item] = True
        formatted_history_items.append(f"{timestamp}  {formatted_item}")
        original_history_items.append(item)

    options = [
        '--info=hidden',
        '--scheme=history',
        '--tiebreak=index',
        '--bind=ctrl-r:up,alt-r:up',
        '--preview-window=down:wrap:nohidden',
        '--no-height',
    ]

    if highlight_preview and which('pygmentize'):
        options.append(f'--preview="printf \'%s\' {{}} | pygmentize -l mysql -P style={shlex.quote(highlight_style)}"')
    else:
        options.append('--preview="printf \'%s\' {}"')

    result = fzf.prompt(
        formatted_history_items,
        fzf_options=' '.join(options),
    )
    safe_invalidate_display(event.app)

    if result:
        selected_index = formatted_history_items.index(result[0])
        buffer.text = original_history_items[selected_index]
        buffer.cursor_position = len(buffer.text)



# type: ignore

import os


def read_fixture_lines(filename):
    """Read lines of text from file.

    :param filename: string name
    :return: list of strings

    """
    lines = []
    for line in open(filename):
        lines.append(line.strip())
    return lines


def read_fixture_files():
    """Read all files inside fixture_data directory."""
    fixture_dict = {}

    current_dir = os.path.dirname(__file__)
    fixture_dir = os.path.join(current_dir, "fixture_data/")
    for filename in os.listdir(fixture_dir):
        if filename not in [".", ".."]:
            fullname = os.path.join(fixture_dir, filename)
            fixture_dict[filename] = read_fixture_lines(fullname)

    return fixture_dict



"""Format adapter for sql."""

from __future__ import annotations

from typing import Generator, Union

from cli_helpers.tabular_output import TabularOutputFormatter

from mycli.packages.sql_utils import extract_tables_from_complete_statements

supported_formats = (
    "sql-insert",
    "sql-update",
    "sql-update-1",
    "sql-update-2",
)

preprocessors = ()

formatter: TabularOutputFormatter


def escape_for_sql_statement(value: Union[bytes, str]) -> str:
    if isinstance(value, bytes):
        return f"0x{value.hex()}"
    else:
        return formatter.mycli.sqlexecute.conn.escape(value)


def adapter(data: list[str], headers: list[str], table_format: Union[str, None] = None, **kwargs) -> Generator[str, None, None]:
    tables = extract_tables_from_complete_statements(formatter.query)
    if len(tables) > 0:
        table = tables[0]
        if table[0]:
            table_name = f'{table[0]}.{table[1]}'
        else:
            table_name = table[1]
    else:
        table_name = "`DUAL`"
    if table_format == "sql-insert":
        h = "`, `".join(headers)
        yield f'INSERT INTO {table_name} (`{h}`) VALUES'
        prefix = "  "
        for d in data:
            values = ", ".join(escape_for_sql_statement(v) for i, v in enumerate(d))
            yield f'{prefix}({values})'
            if prefix == "  ":
                prefix = ", "
        yield ";"
    if table_format and table_format.startswith("sql-update"):
        s = table_format.split("-")
        keys = 1
        if len(s) > 2:
            keys = int(s[-1])
        for d in data:
            yield f'UPDATE {table_name} SET'
            prefix = "  "
            for i, v in enumerate(d[keys:], keys):
                yield f'{prefix}`{headers[i]}` = {escape_for_sql_statement(v)}'
                if prefix == "  ":
                    prefix = ", "
            f = "`{}` = {}"
            where = (f.format(headers[i], escape_for_sql_statement(d[i])) for i in range(keys))
            yield f'WHERE {" AND ".join(where)};'


def register_new_formatter(tof: TabularOutputFormatter):
    global formatter
    formatter = tof
    for sql_format in supported_formats:
        tof.register_new_formatter(sql_format, adapter, preprocessors, {"table_format": sql_format})



# type: ignore

from prompt_toolkit.completion import Completion
from prompt_toolkit.document import Document
import pytest

from test.utils import pygments_below


@pytest.fixture
def completer():
    import mycli.sqlcompleter as sqlcompleter

    return sqlcompleter.SQLCompleter(smart_completion=False)


@pytest.fixture
def complete_event():
    from unittest.mock import Mock

    return Mock()


def test_empty_string_completion(completer, complete_event):
    text = ""
    position = 0
    result = list(completer.get_completions(Document(text=text, cursor_position=position), complete_event))
    assert result == list(map(Completion, completer.all_completions))


def test_select_keyword_completion(completer, complete_event):
    text = "SEL"
    position = len("SEL")
    result = list(completer.get_completions(Document(text=text, cursor_position=position), complete_event))
    assert result == [Completion(text="SELECT", start_position=-3)]


def test_function_name_completion(completer, complete_event):
    text = "SELECT MA"
    position = len("SELECT MA")
    result = list(completer.get_completions(Document(text=text, cursor_position=position), complete_event))
    expected = [
        'MAKEDATE',
        'MAKETIME',
        'MAKE_SET',
        'MANUAL',
        'MASTER',
        'MASTER_AUTO_POSITION',
        'MASTER_BIND',
        'MASTER_COMPRESSION_ALGORITHMS',
        'MASTER_CONNECT_RETRY',
        'MASTER_DELAY',
        'MASTER_HEARTBEAT_PERIOD',
        'MASTER_HOST',
        'MASTER_LOG_FILE',
        'MASTER_LOG_POS',
        'MASTER_PASSWORD',
        'MASTER_PORT',
        'MASTER_POS_WAIT',
        'MASTER_PUBLIC_KEY_PATH',
        'MASTER_RETRY_COUNT',
        'MASTER_SSL',
        'MASTER_SSL_CA',
        'MASTER_SSL_CAPATH',
        'MASTER_SSL_CERT',
        'MASTER_SSL_CIPHER',
        'MASTER_SSL_CRL',
        'MASTER_SSL_CRLPATH',
        'MASTER_SSL_KEY',
        'MASTER_SSL_VERIFY_SERVER_CERT',
        'MASTER_TLS_CIPHERSUITES',
        'MASTER_TLS_VERSION',
        'MASTER_USER',
        'MASTER_ZSTD_COMPRESSION_LEVEL',
        'MATCH',
        'MAX',
        'MAXVALUE',
        'MAX_CONNECTIONS_PER_HOUR',
        'MAX_QUERIES_PER_HOUR',
        'MAX_ROWS',
        'MAX_SIZE',
        'MAX_UPDATES_PER_HOUR',
        'MAX_USER_CONNECTIONS',
    ]

    if pygments_below("2.20"):
        expected.remove('MANUAL')

    assert sorted(x.text for x in result) == sorted(expected)


def test_column_name_completion(completer, complete_event):
    text = "SELECT  FROM users"
    position = len("SELECT ")
    result = list(completer.get_completions(Document(text=text, cursor_position=position), complete_event))
    assert result == list(map(Completion, completer.all_completions))


def test_special_name_completion(completer, complete_event):
    text = "\\"
    position = len("\\")
    result = set(completer.get_completions(Document(text=text, cursor_position=position), complete_event))
    # Special commands will NOT be suggested during naive completion mode.
    assert result == set()



from __future__ import annotations

import re
from typing import Any, Generator, Literal

import sqlglot
import sqlglot.tokens
import sqlparse
from sqlparse.sql import Function, Identifier, IdentifierList, Token, TokenList
from sqlparse.tokens import DML, Keyword, Punctuation

sqlparse.engine.grouping.MAX_GROUPING_DEPTH = None  # type: ignore[assignment]
sqlparse.engine.grouping.MAX_GROUPING_TOKENS = None  # type: ignore[assignment]

cleanup_regex: dict[str, re.Pattern] = {
    # This matches only alphanumerics and underscores.
    "alphanum_underscore": re.compile(r"(\w+)$"),
    # This matches everything except spaces, parens, colon, and comma
    "many_punctuations": re.compile(r"([^():,\s]+)$"),
    # This matches everything except spaces, parens, colon, comma, and period
    "most_punctuations": re.compile(r"([^\.():,\s]+)$"),
    # This matches everything except a space.
    "all_punctuations": re.compile(r"([^\s]+)$"),
}


def last_word(
    text: str,
    include: Literal[
        'alphanum_underscore',
        'many_punctuations',
        'most_punctuations',
        'all_punctuations',
    ] = 'alphanum_underscore',
) -> str:
    r"""
    Find the last word in a sentence.

    >>> last_word('abc')
    'abc'
    >>> last_word(' abc')
    'abc'
    >>> last_word('')
    ''
    >>> last_word(' ')
    ''
    >>> last_word('abc ')
    ''
    >>> last_word('abc def')
    'def'
    >>> last_word('abc def ')
    ''
    >>> last_word('abc def;')
    ''
    >>> last_word('bac $def')
    'def'
    >>> last_word('bac $def', include='most_punctuations')
    '$def'
    >>> last_word('bac \def', include='most_punctuations')
    '\\\\def'
    >>> last_word('bac \def;', include='most_punctuations')
    '\\\\def;'
    >>> last_word('bac::def', include='most_punctuations')
    'def'
    """

    if not text:  # Empty string
        return ""

    if text[-1].isspace():
        return ""
    else:
        regex = cleanup_regex[include]
        matches = regex.search(text)
        if matches:
            return matches.group(0)
        else:
            return ""


# This code is borrowed from sqlparse example script.
# <url>
def is_subselect(parsed: TokenList) -> bool:
    if not parsed.is_group:
        return False
    for item in parsed.tokens:
        if item.ttype is DML and item.value.upper() in ("SELECT", "INSERT", "UPDATE", "CREATE", "DELETE"):
            return True
    return False


def get_last_select(parsed: TokenList) -> TokenList:
    """
    Takes a parsed sql statement and returns the last select query where applicable.

    The intended use case is for when giving table suggestions based on columns, where
    we only want to look at the columns from the most recent select. This works for a single
    select query, or one or more sub queries (the useful part).

    The custom logic is necessary because the typical sqlparse logic for things like finding
    sub selects (i.e. is_subselect) only works on complete statements, such as:

    * select c1 from t1;

    However when suggesting tables based on columns, we only have partial select statements, i.e.:

    * select c1
    * select c1 from (select c2)

    So given the above, we must parse them ourselves as they are not viewed as complete statements.

    Returns a TokenList of the last select statement's tokens.
    """
    select_indexes: list[int] = []

    for token in parsed:
        if token.match(DML, "select"):  # match is case insensitive
            select_indexes.append(parsed.token_index(token))

    last_select = TokenList()

    if select_indexes:
        last_select = TokenList(parsed[select_indexes[-1] :])

    return last_select


def extract_from_part(parsed: TokenList, stop_at_punctuation: bool = True) -> Generator[Any, None, None]:
    tbl_prefix_seen = False
    for item in parsed.tokens:
        if tbl_prefix_seen:
            if is_subselect(item):
                yield from extract_from_part(item, stop_at_punctuation)
            elif stop_at_punctuation and item.ttype is Punctuation:
                return None
            # Multiple JOINs in the same query won't work properly since
            # "ON" is a keyword and will trigger the next elif condition.
            # So instead of stooping the loop when finding an "ON" skip it
            # eg: 'SELECT * FROM abc JOIN def ON abc.id = def.abc_id JOIN ghi'
            elif item.ttype is Keyword and item.value.upper() == "ON":
                tbl_prefix_seen = False
                continue
            # An incomplete nested select won't be recognized correctly as a
            # sub-select. eg: 'SELECT * FROM (SELECT id FROM user'. This causes
            # the second FROM to trigger this elif condition resulting in a
            # StopIteration. So we need to ignore the keyword if the keyword
            # FROM.
            # Also 'SELECT * FROM abc JOIN def' will trigger this elif
            # condition. So we need to ignore the keyword JOIN and its variants
            # INNER JOIN, FULL OUTER JOIN, etc.
            elif item.ttype is Keyword and (not item.value.upper() == "FROM") and (not item.value.upper().endswith("JOIN")):
                return None
            else:
                yield item
        elif (item.ttype is Keyword or item.ttype is Keyword.DML) and item.value.upper() in (
            "COPY",
            "FROM",
            "INTO",
            "UPDATE",
            "TABLE",
            "JOIN",
        ):
            tbl_prefix_seen = True
        # 'SELECT a, FROM abc' will detect FROM as part of the column list.
        # So this check here is necessary.
        elif isinstance(item, IdentifierList):
            for identifier in item.get_identifiers():
                if identifier.ttype is Keyword and identifier.value.upper() == "FROM":
                    tbl_prefix_seen = True
                    break


def extract_table_identifiers(token_stream: Generator[Any, None, None]) -> Generator[tuple[str | None, str, str], None, None]:
    """yields tuples of (schema_name, table_name, table_alias)"""

    for item in token_stream:
        if isinstance(item, IdentifierList):
            for identifier in item.get_identifiers():
                # Sometimes Keywords (such as FROM ) are classified as
                # identifiers which don't have the get_real_name() method.
                try:
                    schema_name = identifier.get_parent_name()
                    real_name = identifier.get_real_name()
                except AttributeError:
                    continue
                if real_name:
                    yield (schema_name, real_name, identifier.get_alias())
        elif isinstance(item, Identifier):
            real_name = item.get_real_name()
            schema_name = item.get_parent_name()

            if real_name:
                yield (schema_name, real_name, item.get_alias())
            else:
                name = item.get_name()
                yield (None, name, item.get_alias() or name)
        elif isinstance(item, Function):
            yield (None, item.get_name(), item.get_name())


# extract_tables is inspired from examples in the sqlparse lib.
def extract_tables(sql: str) -> list[tuple[str | None, str, str]]:
    """Extract the table names from an SQL statement.

    Returns a list of (schema, table, alias) tuples

    """
    parsed = sqlparse.parse(sql)
    if not parsed:
        return []

    # INSERT statements must stop looking for tables at the sign of first
    # Punctuation. eg: INSERT INTO abc (col1, col2) VALUES (1, 2)
    # abc is the table name, but if we don't stop at the first lparen, then
    # we'll identify abc, col1 and col2 as table names.
    insert_stmt = parsed[0].token_first().value.lower() == "insert"
    stream = extract_from_part(parsed[0], stop_at_punctuation=insert_stmt)
    return list(extract_table_identifiers(stream))


def extract_columns_from_select(sql: str) -> list[str]:
    """
    Extract the column names from a select SQL statement.

    Returns a list of columns.
    """
    parsed = sqlparse.parse(sql)
    if not parsed:
        return []

    statement = get_last_select(parsed[0])

    # if there is no select, skip checking for columns
    if not statement:
        return []

    columns = []

    # Loops through the tokens (pieces) of the SQL statement.
    # Once it finds the SELECT token (generally first), it
    # will then start looking for columns from that point on.
    # The get_real_name() function returns the real column name
    # even if an alias is used.
    found_select = False
    for token in statement.tokens:
        if token.ttype is DML and token.value.upper() == 'SELECT':
            found_select = True
        elif found_select:
            if isinstance(token, IdentifierList):
                # multiple columns
                for identifier in token.get_identifiers():
                    if isinstance(identifier, Identifier):
                        column = identifier.get_real_name()
                    elif isinstance(identifier, Token):
                        column = identifier.value
                    else:
                        continue
                    columns.append(column)
            elif isinstance(token, Identifier):
                # single column
                column = token.get_real_name()
                columns.append(column)
            elif token.ttype is Keyword:
                break

            if columns:
                break
    return columns


def extract_tables_from_complete_statements(sql: str) -> list[tuple[str | None, str, str | None]]:
    """Extract the table names from a complete and valid series of SQL
    statements.

    Returns a list of (schema, table, alias) tuples

    """
    # sqlglot chokes entirely on things like "\T" that it doesn't know about,
    # but is much better at extracting table names from complete statements.
    # sqlparse can extract the series of statements, though it also doesn't
    # understand "\T".
    roughly_parsed = sqlparse.parse(sql)
    if not roughly_parsed:
        return []

    finely_parsed = []
    for rough_statement in roughly_parsed:
        try:
            finely_parsed.append(sqlglot.parse_one(str(rough_statement), read='mysql'))
        except sqlglot.errors.ParseError:
            pass

    tables = []
    for fine_statement in finely_parsed:
        for identifier in fine_statement.find_all(sqlglot.exp.Table):
            if identifier.parent_select and identifier.parent_select.sql().startswith('WITH'):
                continue
            tables.append((
                None if identifier.db == '' else identifier.db,
                identifier.name,
                None if identifier.alias == '' else identifier.alias,
            ))

    return tables


def find_prev_keyword(sql: str) -> tuple[Token | None, str]:
    """Find the last sql keyword in an SQL statement

    Returns the value of the last keyword, and the text of the query with
    everything after the last keyword stripped
    """
    if not sql.strip():
        return None, ""

    parsed = sqlparse.parse(sql)[0]
    flattened = list(parsed.flatten())

    logical_operators = ("AND", "OR", "NOT", "BETWEEN")

    for t in reversed(flattened):
        if t.value == "(" or (t.is_keyword and (t.value.upper() not in logical_operators)):
            # Find the location of token t in the original parsed statement
            # We can't use parsed.token_index(t) because t may be a child token
            # inside a TokenList, in which case token_index thows an error
            # Minimal example:
            #   p = sqlparse.parse('select * from foo where bar')
            #   t = list(p.flatten())[-3]  # The "Where" token
            #   p.token_index(t)  # Throws ValueError: not in list
            idx = flattened.index(t)

            # Combine the string values of all tokens in the original list
            # up to and including the target keyword token t, to produce a
            # query string with everything after the keyword token removed
            text = "".join(tok.value for tok in flattened[: idx + 1])
            return t, text

    return None, ""


def query_starts_with(query: str, prefixes: list[str]) -> bool:
    """Check if the query starts with any item from *prefixes*."""
    prefixes = [prefix.lower() for prefix in prefixes]
    formatted_sql = sqlparse.format(query.lower(), strip_comments=True)
    return bool(formatted_sql) and formatted_sql.split()[0] in prefixes


def queries_start_with(queries: str, prefixes: list[str]) -> bool:
    """Check if any queries start with any item from *prefixes*."""
    for query in sqlparse.split(queries):
        if query and query_starts_with(query, prefixes) is True:
            return True
    return False


def query_has_where_clause(query: str) -> bool:
    """Check if the query contains a where-clause."""
    return any(isinstance(token, sqlparse.sql.Where) for token_list in sqlparse.parse(query) for token in token_list)


# todo: handle "UPDATE LOW_PRIORITY" and "UPDATE IGNORE"
def query_is_single_table_update(query: str) -> bool:
    """Check if a query is a simple single-table UPDATE."""
    cleaned_query_for_parsing_only = sqlparse.format(query, strip_comments=True)
    cleaned_query_for_parsing_only = re.sub(r'\s+', ' ', cleaned_query_for_parsing_only)
    if not cleaned_query_for_parsing_only:
        return False
    parsed = sqlparse.parse(cleaned_query_for_parsing_only)
    if not parsed:
        return False
    statement = parsed[0]
    try:
        retval = bool(
            statement[0].value.lower() == 'update'
            and statement[1].is_whitespace
            and ',' not in statement[2].value  # multiple tables
            and statement[3].is_whitespace
            and statement[4].value.lower() == 'set'
        )
    except IndexError:
        retval = False

    return retval


def is_destructive(keywords: list[str], queries: str) -> bool:
    """Returns True if any of the queries in *queries* is destructive."""
    for query in sqlparse.split(queries):
        if not query:
            continue
        # subtle: if "UPDATE" is one of our keywords AND "query" starts with "UPDATE"
        if query_starts_with(query, keywords) and query_starts_with(query, ["update"]):
            if query_has_where_clause(query) and query_is_single_table_update(query):
                return False
            else:
                return True
        if query_starts_with(query, keywords):
            return True

    return False


def is_dropping_database(queries: str, dbname: str | None) -> bool:
    """Determine if the query is dropping a specific database."""
    result = False
    if dbname is None:
        return False

    def normalize_db_name(db: str) -> str:
        return db.lower().strip('`"')

    dbname = normalize_db_name(dbname)

    for query in sqlparse.parse(queries):
        keywords = [t for t in query.tokens if t.is_keyword]
        if len(keywords) < 2:
            continue
        if keywords[0].normalized in ("DROP", "CREATE") and keywords[1].value.lower() in (
            "database",
            "schema",
        ):
            database_token = next((t for t in query.tokens if isinstance(t, Identifier)), None)
            if database_token is not None and normalize_db_name(database_token.get_name()) == dbname:
                result = keywords[0].normalized == "DROP"
    return result


def need_completion_refresh(queries: str) -> bool:
    """Determines if the completion needs a refresh by checking if the sql
    statement is an alter, create, drop or change db."""
    for query in sqlparse.split(queries):
        try:
            first_token = query.split()[0]
            if first_token.lower() in ("alter", "create", "use", "\\r", "\\u", "connect", "drop", "rename"):
                return True
        except Exception:
            continue
    return False


def need_completion_reset(queries: str) -> bool:
    """Determines if the statement is a database switch such as 'use' or '\\u'.
    When a database is changed the existing completions must be reset before we
    start the completion refresh for the new database.
    """
    for query in sqlparse.split(queries):
        try:
            tokens = query.split()
            first_token = tokens[0]
            if first_token.lower() in ("use", "\\u"):
                return True
            if first_token.lower() in ("\\r", "connect") and len(tokens) > 1:
                return True
        except Exception:
            continue
    return False


def is_mutating(status_plain: str | None) -> bool:
    """Determines if the statement is mutating based on the status."""
    if not status_plain:
        return False

    mutating = {"insert", "update", "delete", "alter", "create", "drop", "replace", "truncate", "load", "rename"}
    return status_plain.split(None, 1)[0].lower() in mutating


def is_select(status_plain: str | None) -> bool:
    """Returns true if the first word in status is 'select'."""
    if not status_plain:
        return False
    return status_plain.split(None, 1)[0].lower() == "select"


def classify_sandbox_statement(text: str) -> tuple[str | None, str | None]:
    """Classify a SQL statement for sandbox mode and extract the new password.

    Returns (statement_type, new_password) where statement_type is one of:
    - 'alter_user'    — ALTER USER ... IDENTIFIED BY ...
    - 'set_password'  — SET PASSWORD [FOR ...] = ...
    - 'quit'          — quit, exit, \\q
    - None            — not allowed in sandbox mode
    """
    stripped = text.strip()
    if not stripped:
        return ('quit', None)

    try:
        tokens = list(sqlglot.tokenize(stripped, dialect='mysql'))
    except sqlglot.errors.TokenError:
        tokens = []

    if not tokens:
        return ('quit', None)

    types = [t.token_type for t in tokens]
    texts = [t.text.upper() for t in tokens]
    tt = sqlglot.tokens.TokenType

    # quit, exit
    if len(tokens) == 1 and types[0] == tt.VAR and texts[0] in ('QUIT', 'EXIT'):
        return ('quit', None)

    # \q
    if len(tokens) == 2 and types[0] == tt.BACKSLASH and texts[1] == 'Q':
        return ('quit', None)

    # ALTER USER ...
    if len(tokens) >= 2 and types[0] == tt.ALTER and texts[1] == 'USER':
        pw = _find_password_after_by(tokens)
        return ('alter_user', pw)

    # SET PASSWORD ...
    if len(tokens) >= 2 and types[0] == tt.SET and texts[1] == 'PASSWORD':
        pw = _find_password_after_eq(tokens)
        return ('set_password', pw)

    return (None, None)


def _find_password_after_by(tokens: list[sqlglot.tokens.Token]) -> str | None:
    """Find a password literal following a BY token (for ALTER USER ... IDENTIFIED BY 'pw')."""
    tt = sqlglot.tokens.TokenType
    for i, tok in enumerate(tokens):
        if tok.token_type == tt.VAR and tok.text.upper() == 'BY' and i + 1 < len(tokens):
            next_tok = tokens[i + 1]
            if next_tok.token_type == tt.STRING:
                return next_tok.text
    return None


def _find_password_after_eq(tokens: list[sqlglot.tokens.Token]) -> str | None:
    """Find a password literal following an = token (for SET PASSWORD = 'pw')."""
    tt = sqlglot.tokens.TokenType
    for i, tok in enumerate(tokens):
        if tok.token_type == tt.EQ and i + 1 < len(tokens):
            next_tok = tokens[i + 1]
            if next_tok.token_type == tt.STRING:
                return next_tok.text
    return None


def is_sandbox_allowed(text: str) -> bool:
    """Return True if the command is allowed in expired-password sandbox mode."""
    stmt_type, _ = classify_sandbox_statement(text)
    return stmt_type is not None


def is_password_change(text: str) -> bool:
    """Return True if the command is a password change statement."""
    stmt_type, _ = classify_sandbox_statement(text)
    return stmt_type in ('alter_user', 'set_password')


def extract_new_password(text: str) -> str | None:
    """Extract the new password from an ALTER USER or SET PASSWORD statement."""
    _, password = classify_sandbox_statement(text)
    return password
